/*

	jackmixdesk_gtk.c
	Copyright (C) 2007  Ulrich Lorenz Schlueter
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>
#include <stringprep.h>
#include <gtk/gtk.h>
#include <phat/phat.h>
#include <lo/lo.h>

#include "db.h"

#define		DEFAULT_CLIENT_NAME		"JackMixDesk_gtk"		/* Default name of client */
#define		DEFAULT_MIXER_PORT		"16000"					/* Default jackmixdesk port */
#define		DEFAULT_MIXER_HOST		"localhost"				/* Default jackmixdesk host */
#define		DEFAULT_LABEL_LEN		(16)					/* Max length of label strings */
#define		DEFAULT_ENTRY_LEN		(4)						/* Max length of entry */
#define		KNOB_GAIN_MIN			(-90)					/* Min range for gain knobs */
#define		KNOB_GAIN_MAX			(0)						/* Max range for gain knobs */
#define		KNOB_GAIN_STEP			(0.05)					/* Step width for gain knobs */
#define		KNOB_PAN_MIN			(-90)					/* Min range for pan knobs */
#define		KNOB_PAN_MAX			(90)					/* Max range for pan knobs */
#define		KNOB_PAN_STEP			(0.05)					/* Step width for pan knobs */
#define		SPINBUTTON_FONT_SIZE	(8400)					/* Font size for spin buttons */
#define		SMALL_LABEL_FONT_SIZE	(8700)					/* Font size for small labels */
#define		MEDIUM_LABEL_FONT_SIZE	(8750)					/* Font size for bigger labels */
#define		KNOB_SIZE				(50)					/* Default knob size */

typedef struct {
	char label[DEFAULT_LABEL_LEN];		/* Label for Pre/Postfader */
	double gain;						/* decibels */
	double pan;
	double gain_l;						/* decibels */
	double gain_r;						/* decibels */
	
	GtkWidget *label_send;
	GtkWidget *box_vbox;
	GtkWidget *box_hbox_upper;
	GtkWidget *box_hbox_middle;
	GtkWidget *box_hbox_lower;
	GtkWidget *box_vbox_middle_left;
	GtkWidget *box_vbox_middle_right;
	GtkWidget *box_hbox_middle_knob_left;
	GtkWidget *box_hbox_middle_knob_right;
	GtkWidget *box_hbox_middle_entry_left;
	GtkWidget *box_hbox_middle_entry_right;
	GtkWidget *box_vbox_lower_left;
	GtkWidget *box_vbox_lower_right;
	GtkWidget *box_hbox_lower_knob_left;
	GtkWidget *box_hbox_lower_knob_right;
	GtkWidget *box_hbox_lower_entry_left;
	GtkWidget *box_hbox_lower_entry_right;
	GtkWidget *label_gain;
	GtkWidget *label_pan;
	GtkWidget *label_gain_l;
	GtkWidget *label_gain_r;
	GtkWidget *knob_gain;
	GtkWidget *knob_pan;
	GtkWidget *knob_gain_l;
	GtkWidget *knob_gain_r;
	GtkWidget *entry_gain;
	GtkWidget *entry_pan;
	GtkWidget *entry_gain_l;
	GtkWidget *entry_gain_r;
	GtkWidget *separator;
} jmd_send_t;

typedef struct {
	char label[DEFAULT_LABEL_LEN];	/* Label for Channel */
	unsigned int button_stereo_state;
	unsigned int button_mute_state;
	
	jmd_send_t *sends;				/* Aux Sends */
	
	GtkWidget *box_vbox;
	GtkWidget *box_hbox_knob_pan;
	GtkWidget *box_hbox_entry_pan;
	GtkWidget *box_hbox_entry_gain;
	GtkWidget *box_hbox_button_stereo;
	GtkWidget *box_hbox_button_mute;
	GtkWidget *label_gain;
	GtkWidget *label_pan;
	GtkWidget *knob_pan;			/* Same as first send gain */
	GtkWidget *fader_gain;			/* Same as first send gain */
	GtkWidget *entry_gain;
	GtkWidget *entry_pan;
	GtkWidget *label_stereo;
	GtkWidget *button_stereo;
	GtkWidget *button_mute;
	GtkWidget *label_channel_top;
	GtkWidget *label_channel_bottom;
	GtkWidget *separator;
	GtkWidget *separator_top;
	GtkWidget *separator_bottom;
} jmd_channel_t;

typedef struct {
	char label[DEFAULT_LABEL_LEN];	/* Label for Output */
	double gain;					/* decibels */
	unsigned int button_mute_state;
	unsigned int button_link_state;
	
	GtkWidget *box_vbox;
	GtkWidget *separator;
	GtkWidget *separator_top;
	GtkWidget *separator_bottom;
	GtkWidget *box_hbox_label_top;
	GtkWidget *box_hbox_label_bottom;
	GtkWidget *box_hbox_fader;
	GtkWidget *box_hbox_entry;
	GtkWidget *label_gain;
	GtkWidget *label_out_bottom;
	GtkWidget *fader_gain;
	GtkWidget *entry_gain;
	GtkWidget *box_hbox_mute;
	GtkWidget *button_mute;
	GtkWidget *box_hbox_link;
	GtkWidget *label_link;
	GtkWidget *button_link;
} jmd_output_t;

typedef struct {
	unsigned int verbose;
	unsigned int quiet;
	unsigned int prefader_count;
	unsigned int postfader_count;
	unsigned int channel_count;
	unsigned int omit_send_pan;
	unsigned int omit_send_gain;
	unsigned int omit_chan_pan;
	char* osc_mixer_port;
	char* osc_mixer_host;
	char client_name[DEFAULT_LABEL_LEN];
	
	jmd_channel_t *channels;
	jmd_output_t *outputs;
	
	lo_server_thread st;
	lo_server serv;
	lo_address t;
	lo_timetag ts;
	
	GtkWidget *window_output;
    GtkWidget *window_outputs;
	GtkWidget *box_outputs_vbox;
	GtkWidget *box_outputs_hbox;
	GtkWidget *menu_bar;
	GtkWidget *file_menu;
	GtkWidget *file_item;
	GtkWidget *help_menu;
	GtkWidget *help_item;
    GtkWidget *quit_item;
    GtkWidget *about_item;
	GtkWidget *window_channel_1;
    GtkWidget *window_channels_1;
	GtkWidget *box_channels_1_hbox;
	GtkWidget *window_channel_2;
    GtkWidget *window_channels_2;
	GtkWidget *box_channels_2_hbox;
	
	GtkStyle *spinbutton_style;
	GtkStyle *small_label_style;
	GtkStyle *medium_label_style;
} jmd_interface_t;

unsigned int running=1;

static void gtk_fader_gain_out_change(GtkWidget *widget, gpointer data)
{
	int i=atoi(gtk_widget_get_name(widget));
	double gain=phat_fan_slider_get_value(PHAT_FAN_SLIDER(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;
	
	if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->outputs[i-1].entry_gain)) != gain ){
		gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->outputs[i-1].entry_gain),gain);
	}
	if ( interface->outputs[i-1].button_link_state == 1 && phat_fan_slider_get_value(PHAT_FAN_SLIDER(interface->outputs[i].fader_gain))!= gain ){
		phat_fan_slider_set_value(PHAT_FAN_SLIDER(interface->outputs[i].fader_gain),gain);
	}
	if ( interface->outputs[i-2].button_link_state == 1 && phat_fan_slider_get_value(PHAT_FAN_SLIDER(interface->outputs[i-2].fader_gain))!= gain ){
		phat_fan_slider_set_value(PHAT_FAN_SLIDER(interface->outputs[i-2].fader_gain),gain);
	}
	
	if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/output/gain_set", "if",i,gain) == -1 ){
		printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
	}
	
	if (interface->verbose) {
	    g_print ("Gtk out gain fader changed\n");
		g_print ("value=%f\n",phat_fan_slider_get_value(PHAT_FAN_SLIDER(widget)));
		g_print ("%i\n",atoi(gtk_widget_get_name(widget)));
	}
}

static void gtk_entry_gain_out_change(GtkWidget *widget, gpointer data)
{
	int i=atoi(gtk_widget_get_name(widget));
	double gain=gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;
	
	if ( phat_fan_slider_get_value(PHAT_FAN_SLIDER(interface->outputs[i-1].fader_gain))!= gain ){
		phat_fan_slider_set_value(PHAT_FAN_SLIDER(interface->outputs[i-1].fader_gain),gain);
	}
	
   	if (interface->verbose) {
		g_print ("Gtk out gain entry changed\n");
		g_print ("value=%f\n",gtk_spin_button_get_value(GTK_SPIN_BUTTON (widget)));
		g_print ("%i\n",atoi(gtk_widget_get_name(widget)));
   	}	
}

static void gtk_button_mute_out_change(GtkWidget *widget, gpointer data)
{
	int i=atoi(gtk_widget_get_name(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;

  	if (interface->verbose) {
    	g_print ("Gtk out %i mute button change\n",i);
    	g_print ("%i\n",interface->outputs[i-1].button_mute_state);
  	}
    
    if (interface->outputs[i-1].button_mute_state==0){
    	interface->outputs[i-1].button_mute_state=1;
    } else {
    	interface->outputs[i-1].button_mute_state=0;
    }
    
    if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/output/mute_set", "ii",i,interface->outputs[i-1].button_mute_state) == -1 ){
		printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
	}
	
	if (interface->verbose) {
		g_print ("Gtk out %i mute button changed\n",i);
		g_print ("%i\n",interface->outputs[i-1].button_mute_state);
	}
}

static void gtk_button_link_out_change(GtkWidget *widget, gpointer data)
{
	int i=atoi(gtk_widget_get_name(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;

  	if (interface->verbose) {
    	g_print ("Gtk out %i link button change\n",i);
    	g_print ("%i\n",interface->outputs[i-1].button_link_state);
  	}
    
    if (interface->outputs[i-1].button_link_state==0){
    	interface->outputs[i-1].button_link_state=1;
    	if ( phat_fan_slider_get_value(PHAT_FAN_SLIDER(interface->outputs[i-1].fader_gain))!= phat_fan_slider_get_value(PHAT_FAN_SLIDER(interface->outputs[i].fader_gain)) ){
			phat_fan_slider_set_value(PHAT_FAN_SLIDER(interface->outputs[i].fader_gain),phat_fan_slider_get_value(PHAT_FAN_SLIDER(interface->outputs[i-1].fader_gain)));
		}
    } else {
    	interface->outputs[i-1].button_link_state=0;
    }
    
    if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/output/link_set", "ii",i,interface->outputs[i-1].button_link_state) == -1 ){
		printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
	}
	
	if (interface->verbose) {
		g_print ("Gtk out %i link button changed\n",i);
		g_print ("%i\n",interface->outputs[i-1].button_link_state);
	}
}

static void gtk_fader_gain_chan_change(GtkWidget *widget, gpointer data)
{
	int i=atoi(gtk_widget_get_name(widget));
	double gain=phat_fan_slider_get_value(PHAT_FAN_SLIDER(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;
	
	if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i-1].entry_gain)) != gain ){
		gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i-1].entry_gain),gain);
	}
	if (phat_knob_get_value(PHAT_KNOB (interface->channels[i-1].sends[0].knob_gain)) != gain ){
		phat_knob_set_value(PHAT_KNOB (interface->channels[i-1].sends[0].knob_gain),gain);
	}
	if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[0].entry_gain)) != gain ){
		gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[0].entry_gain),gain);
	}
	if ( interface->channels[i-1].button_stereo_state == 1 && phat_fan_slider_get_value(PHAT_FAN_SLIDER(interface->channels[i].fader_gain))!= gain ){
		phat_fan_slider_set_value(PHAT_FAN_SLIDER(interface->channels[i].fader_gain),gain);
	}
	if ( interface->channels[i-2].button_stereo_state == 1 && phat_fan_slider_get_value(PHAT_FAN_SLIDER(interface->channels[i-2].fader_gain))!= gain ){
		phat_fan_slider_set_value(PHAT_FAN_SLIDER(interface->channels[i-2].fader_gain),gain);
	}
	
	if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_set", "iif",i,1,gain) == -1 ){
		printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
	}

  	if (interface->verbose) {
	    g_print ("Gtk chan gain fader changed\n");
		g_print ("value=%f\n",phat_fan_slider_get_value(PHAT_FAN_SLIDER(widget)));
		g_print ("%i\n",atoi(gtk_widget_get_name(widget)));
  	}	
}

static void gtk_entry_gain_chan_change(GtkWidget *widget, gpointer data)
{
	int i=atoi(gtk_widget_get_name(widget));
	double gain=gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;
	
	if ( phat_fan_slider_get_value(PHAT_FAN_SLIDER(interface->channels[i-1].fader_gain))!= gain ){
		phat_fan_slider_set_value(PHAT_FAN_SLIDER(interface->channels[i-1].fader_gain),gain);
	}
	
	if (PHAT_IS_KNOB(interface->channels[i-1].sends[0].knob_gain)){
		if (phat_knob_get_value(PHAT_KNOB (interface->channels[i-1].sends[0].knob_gain)) != gain ){
			phat_knob_set_value(PHAT_KNOB (interface->channels[i-1].sends[0].knob_gain),gain);
		}
	}
	
		if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[0].entry_gain)) != gain ){
			gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[0].entry_gain),gain);
		}
	
  	if (interface->verbose) {
	    g_print ("Gtk chan gain entry changed\n");
		g_print ("value=%f\n",gtk_spin_button_get_value(GTK_SPIN_BUTTON (widget)));
		g_print ("%i\n",atoi(gtk_widget_get_name(widget)));
  	}	
}

static void gtk_button_mute_chan_change(GtkWidget *widget, gpointer data)
{
	int i=atoi(gtk_widget_get_name(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;

  	if (interface->verbose) {
    	g_print ("Gtk chan %i mute button change\n",i);
    	g_print ("%i\n",interface->channels[i-1].button_mute_state);
  	}
    
    if (interface->channels[i-1].button_mute_state==0){
    	interface->channels[i-1].button_mute_state=1;
    } else {
    	interface->channels[i-1].button_mute_state=0;
    }
    
    if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/mute_set", "ii",i,interface->channels[i-1].button_mute_state) == -1 ){
		printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
	}
	
	if (interface->verbose) {
		g_print ("Gtk chan %i mute button changed\n",i);
		g_print ("%i\n",interface->channels[i-1].button_mute_state);
	}
}

static void gtk_button_stereo_chan_change(GtkWidget *widget, gpointer data)
{
	int i=atoi(gtk_widget_get_name(widget));
	int s;
	double left=-90;
	double right=90;
	double low=-90;
	double high=0;
	jmd_interface_t *interface = (jmd_interface_t*)data;

  	if (interface->verbose) {
    	g_print ("Gtk chan %i stereo button change\n",i);
    	g_print ("%i\n",interface->channels[i-1].button_stereo_state);
  	}
    
    if (interface->channels[i-1].button_stereo_state==0){
    	interface->channels[i-1].button_stereo_state=1;

		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/stereo_set", "ii",i,1) == -1 ){
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
    	if ( phat_fan_slider_get_value(PHAT_FAN_SLIDER(interface->channels[i-1].fader_gain))!= phat_fan_slider_get_value(PHAT_FAN_SLIDER(interface->channels[i].fader_gain)) ){
			phat_fan_slider_set_value(PHAT_FAN_SLIDER(interface->channels[i].fader_gain),phat_fan_slider_get_value(PHAT_FAN_SLIDER(interface->channels[i-1].fader_gain)));
		}
		if (!interface->omit_chan_pan ){
	    	if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i-1].entry_pan)) != left ){
				gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i-1].entry_pan),left);
	    	}
	    	if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i].entry_pan)) != right ){
				gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i].entry_pan),right);
	    	}
		}
		
	 	for ( s=0;s<(interface->prefader_count+interface->postfader_count);s++){
			if (!interface->omit_send_pan){	 		
				if (phat_knob_get_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_pan)) != left ){
					phat_knob_set_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_pan),left);
				}
				if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_pan)) != left ){
					gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_pan),left);
				}
				if (phat_knob_get_value(PHAT_KNOB (interface->channels[i].sends[s].knob_pan)) != right ){
					phat_knob_set_value(PHAT_KNOB (interface->channels[i].sends[s].knob_pan),right);
				}
				if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i].sends[s].entry_pan)) != right ){
					gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i].sends[s].entry_pan),right);
				}
			}
			if (!interface->omit_send_gain){
				if (phat_knob_get_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_gain_l)) != high ){
					phat_knob_set_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_gain_l),high);
				}
				if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_gain_l)) != high ){
					gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_gain_l),high);
				}
				if (phat_knob_get_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_gain_r)) != low ){
					phat_knob_set_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_gain_r),low);
				}
				if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_gain_r)) != low ){
					gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_gain_r),low);
				}
				if (phat_knob_get_value(PHAT_KNOB (interface->channels[i].sends[s].knob_gain_l)) != low ){
					phat_knob_set_value(PHAT_KNOB (interface->channels[i].sends[s].knob_gain_l),low);
				}
				if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i].sends[s].entry_gain_l)) != low ){
					gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i].sends[s].entry_gain_l),low);
				}
				if (phat_knob_get_value(PHAT_KNOB (interface->channels[i].sends[s].knob_gain_r)) != high ){
					phat_knob_set_value(PHAT_KNOB (interface->channels[i].sends[s].knob_gain_r),high);
				}
				if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i].sends[s].entry_gain_r)) != high ){
					gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i].sends[s].entry_gain_r),high);
				}
			}
			if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_l_set", "iif",i,s+1,high) == -1 ){
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
			if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_r_set", "iif",i,s+1,low) == -1 ){
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
			if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_l_set", "iif",i+1,s+1,low) == -1 ){
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
			if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_r_set", "iif",i+1,s+1,high) == -1 ){
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
		}
    } else {
    	interface->channels[i-1].button_stereo_state=0;
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/stereo_set", "ii",i,0) == -1 ){
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}    
	}
    
	if (interface->verbose) {
		g_print ("Gtk chan %i stereo button changed\n",i);
		g_print ("%i\n",interface->channels[i-1].button_stereo_state);
	}
}

static void gtk_fader_pan_chan_change(GtkWidget *widget, gpointer data)
{
	int i=atoi(gtk_widget_get_name(widget));
	int s;
	double pan=phat_knob_get_value(PHAT_KNOB(widget));
	double gain_l,gain_r;
	jmd_interface_t *interface = (jmd_interface_t*)data;
	
	if (interface->verbose) {	
	    g_print ("Gtk chan pan fader change\n");
		g_print ("value=%f\n",phat_knob_get_value(PHAT_KNOB(widget)));	
		g_print ("%i\n",atoi(gtk_widget_get_name(widget)));
	}
	
	if ( pan < 0 ){
		gain_l = 0;
		gain_r = pan;
	} else if ( pan > 0 ){
		gain_l = pan * (-1);
		gain_r = 0;
	} else {
		gain_l = 0;
		gain_r = 0;
	}
	
	if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i-1].entry_pan)) != pan ){
		gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i-1].entry_pan),pan);
	}
	
	for ( s=0;s<(interface->prefader_count+interface->postfader_count);s++){
		if (!interface->omit_send_pan){
			if (phat_knob_get_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_pan)) != pan ){
				phat_knob_set_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_pan),pan);
			}
			if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_pan)) != pan ){
				gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_pan),pan);
			}
		}
		if (!interface->omit_send_gain){		
			if (phat_knob_get_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_gain_l)) != gain_l ){
				phat_knob_set_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_gain_l),gain_l);
			}
			if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_gain_l)) != gain_l ){
				gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_gain_l),gain_l);
			}
			if (phat_knob_get_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_gain_r)) != gain_r ){
				phat_knob_set_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_gain_r),gain_r);
			}
			if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_gain_r)) != gain_r ){
				gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_gain_r),gain_r);
			}
		}
		
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_l_set", "iif",i,s+1,gain_l) == -1 ){
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_r_set", "iif",i,s+1,gain_r) == -1 ){
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}		
	}
	
	
	if (interface->verbose) {	
	    g_print ("Gtk chan pan fader changed\n");
		g_print ("value=%f\n",phat_knob_get_value(PHAT_KNOB(widget)));	
		g_print ("%i\n",atoi(gtk_widget_get_name(widget)));
	}	
}

static void gtk_entry_pan_chan_change(GtkWidget *widget, gpointer data)
{
	int i=atoi(gtk_widget_get_name(widget));
	double pan=gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;
	int s;

	if (interface->verbose) {	
	    g_print ("Gtk chan pan entry changed\n");
		g_print ("value=%f\n",gtk_spin_button_get_value(GTK_SPIN_BUTTON (widget)));	
		g_print ("%i\n",atoi(gtk_widget_get_name(widget)));	
	}

	if (phat_knob_get_value(PHAT_KNOB (interface->channels[i-1].knob_pan)) != pan ){
		phat_knob_set_value(PHAT_KNOB (interface->channels[i-1].knob_pan),pan);
	}
	
	for (s=0;s<interface->prefader_count+interface->postfader_count;s++){
		if (PHAT_IS_KNOB(interface->channels[i-1].sends[s].knob_pan)){
			if (phat_knob_get_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_pan)) != pan ){
				phat_knob_set_value(PHAT_KNOB (interface->channels[i-1].sends[s].knob_pan),pan);
			}
		}
		if (GTK_IS_SPIN_BUTTON(interface->channels[i-1].sends[s].entry_pan)){
			if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_pan)) != pan ){
				gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[i-1].sends[s].entry_pan),pan);
			}
		}
	}
	
	if (interface->verbose) {	
	    g_print ("Gtk chan pan entry changed\n");
		g_print ("value=%f\n",gtk_spin_button_get_value(GTK_SPIN_BUTTON (widget)));	
		g_print ("%i\n",atoi(gtk_widget_get_name(widget)));	
	}	
}

static void gtk_fader_gain_send_change(GtkWidget *widget, gpointer data)
{
	int i,h,c,s;
	char chan[10];
	char send[10];
	char name[10];
	double gain=phat_knob_get_value(PHAT_KNOB(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;
	
	sprintf(name,"%s",gtk_widget_get_name(widget));
	
	i=0;
	h=0;
	while (name[i] != ','){
		chan[i] = name[i];
		i++;
	}
	if (name[i] == ','){
		chan[i] = '\0';
		i++;
	}
	while (name[i] != '\0'){
		send[h] = name[i];
		i++;
		h++;
	}
	if (name[i] == '\0'){
		send[h] = '\0';
	}

	c=atoi(chan);
	s=atoi(send);

	if(s == 1) phat_fan_slider_set_value(PHAT_FAN_SLIDER(interface->channels[c-1].fader_gain),gain);
	
	if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain)) != gain ){
		gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain),gain);
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_set", "iif",c,s,gain) == -1 ){
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
	}
	if ( interface->channels[c-1].button_stereo_state == 1 && gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[c].sends[s-1].entry_gain))!= gain ){
		gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[c].sends[s-1].entry_gain),gain);
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_set", "iif",c+1,s,gain) == -1 ){
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}		
	}
	if ( interface->channels[c-2].button_stereo_state == 1 && gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[c-2].sends[s-1].entry_gain))!= gain ){
		gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[c-2].sends[s-1].entry_gain),gain);
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_set", "iif",c-1,s,gain) == -1 ){
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
	}
	
	if (interface->verbose) {	
	    g_print ("Gtk send gain fader changed\n");
		g_print ("name=%s\n",gtk_widget_get_name(widget));
		g_print ("value=%f\n",phat_knob_get_value(PHAT_KNOB(widget)));
		g_print ("channel=%i send=%i\n",c,s);	
	}	
}

static void gtk_entry_gain_send_change(GtkWidget *widget, gpointer data)
{
	int i,h,c,s;
	char chan[10];
	char send[10];
	char name[10];
	double gain=gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;

	sprintf(name,"%s",gtk_widget_get_name(widget));
	
	i=0;
	h=0;
	while (name[i] != ','){
		chan[i] = name[i];
		i++;
	}
	if (name[i] == ','){
		chan[i] = '\0';
		i++;
	}
	while (name[i] != '\0'){
		send[h] = name[i];
		i++;
		h++;
	}
	if (name[i] == '\0'){
		send[h] = '\0';
	}

	c=atoi(chan);
	s=atoi(send);
	
	if (phat_knob_get_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain)) != gain ){
		phat_knob_set_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain),gain);
	}
	if(s == 1){
		if ( phat_fan_slider_get_value(PHAT_FAN_SLIDER(interface->channels[c-1].fader_gain))!= gain ){
			phat_fan_slider_set_value(PHAT_FAN_SLIDER(interface->channels[c-1].fader_gain),gain);
		}
	}		
	if (interface->verbose) {	
	    g_print ("Gtk send gain entry changed\n");
		g_print ("name=%s\n",gtk_widget_get_name(widget));
		g_print ("value=%f\n",gtk_spin_button_get_value(GTK_SPIN_BUTTON (widget)));	
		g_print ("channel=%i send=%i\n",c,s);	
	}	
}

static void gtk_fader_pan_send_change(GtkWidget *widget, gpointer data)
{
	int i,h,c,s;
	char chan[10];
	char send[10];
	char name[10];
	double gain_l,gain_r;
	double pan=phat_knob_get_value(PHAT_KNOB (widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;

	sprintf(name,"%s",gtk_widget_get_name(widget));
	
	if ( pan < 0 ){
		gain_l = 0;
		gain_r = pan;
	} else if ( pan > 0 ){
		gain_l = pan * (-1);
		gain_r = 0;
	} else {
		gain_l = 0;
		gain_r = 0;
	}
	
	i=0;
	h=0;
	while (name[i] != ','){
		chan[i] = name[i];
		i++;
	}
	if (name[i] == ','){
		chan[i] = '\0';
		i++;
	}
	while (name[i] != '\0'){
		send[h] = name[i];
		i++;
		h++;
	}
	if (name[i] == '\0'){
		send[h] = '\0';
	}
	
	c=atoi(chan);
	s=atoi(send);
	
	if (!interface->omit_send_pan){
		if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_pan)) != pan ){
			gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_pan),pan);
		}
	}
	
	if (!interface->omit_send_gain){		
		if (phat_knob_get_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_l)) != gain_l ){
			phat_knob_set_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_l),gain_l);
		}
		if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_l)) != gain_l ){
			gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_l),gain_l);
		}
		if (phat_knob_get_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_r)) != gain_r ){
			phat_knob_set_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_r),gain_r);
		}
		if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_r)) != gain_r ){
			gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_r),gain_r);
		}
	}
	
	if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_l_set", "iif",c,s,gain_l) == -1 ){
		printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
	}
	if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_r_set", "iif",c,s,gain_r) == -1 ){
		printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
	}
				
	if (interface->verbose) {	
	    g_print ("Gtk send pan fader changed\n");
		g_print ("name=%s\n",gtk_widget_get_name(widget));
		g_print ("value=%f\n",phat_knob_get_value(PHAT_KNOB(widget)));	
		g_print ("channel=%i send=%i\n",c,s);	
	}	
}

static void gtk_entry_pan_send_change(GtkWidget *widget, gpointer data)
{
	int i,h,c,s;
	char chan[10];
	char send[10];
	char name[10];
	double pan=gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;
	
	sprintf(name,"%s",gtk_widget_get_name(widget));
	
	i=0;
	h=0;
	while (name[i] != ','){
		chan[i] = name[i];
		i++;
	}
	if (name[i] == ','){
		chan[i] = '\0';
		i++;
	}
	while (name[i] != '\0'){
		send[h] = name[i];
		i++;
		h++;
	}
	if (name[i] == '\0'){
		send[h] = '\0';
	}

	c=atoi(chan);
	s=atoi(send);

	if(s-1 == 0){
		if (GTK_IS_SPIN_BUTTON(interface->channels[c-1].entry_pan)){
			if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[c-1].entry_pan)) != pan ){
				gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[c-1].entry_pan),pan);
			}
		}
		if (PHAT_IS_KNOB(interface->channels[c-1].knob_pan)){
			if (phat_knob_get_value(PHAT_KNOB (interface->channels[c-1].knob_pan)) != pan ){
				phat_knob_set_value(PHAT_KNOB (interface->channels[c-1].knob_pan),pan);
			}
		}
	}

	if (PHAT_IS_KNOB(interface->channels[c-1].sends[s-1].knob_pan)){
		if (phat_knob_get_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_pan)) != pan ){
			phat_knob_set_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_pan),pan);
		}
	}

	if (PHAT_IS_KNOB(interface->channels[c-1].sends[s-1].knob_gain_l) && PHAT_IS_KNOB(interface->channels[c-1].sends[s-1].knob_gain_r)){	
		if (pan <= 0){
			if (phat_knob_get_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_l)) != 0 ){
				phat_knob_set_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_l),0);
			}
			if (phat_knob_get_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_r)) != pan ){
				phat_knob_set_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_r),pan);
			}
		}
		if (pan >= 0){
			if (phat_knob_get_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_l)) != pan  * (-1) ){
				phat_knob_set_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_l),pan  * (-1));
			}
			if (phat_knob_get_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_r)) !=  0){
				phat_knob_set_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_r),0);
			}
		}		
	}
	if (GTK_IS_SPIN_BUTTON(interface->channels[c-1].sends[s-1].entry_gain_l) && GTK_IS_SPIN_BUTTON(interface->channels[c-1].sends[s-1].entry_gain_r)){
		if (pan <= 0){
			if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_l)) != 0 ){
				gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_l),0);
			}
			if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_r)) != pan ){
				gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_r),pan);
			}
		} else if (pan >= 0){
			if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_l)) != pan  * (-1) ){
				gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_l),pan  * (-1));
			}
			if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_r)) != 0){
				gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_r),0);
			}
		}
	}
	
	if (pan <= 0){
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_l_set", "iif",c,s,0.0) == -1 ){
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_r_set", "iif",c,s,pan) == -1 ){
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
	} else {
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_l_set", "iif",c,s,pan * (-1)) == -1 ){
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_r_set", "iif",c,s,0.0) == -1 ){
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
	}		
					
	if (interface->verbose) {	
	    g_print ("Gtk send pan entry changed\n");
		g_print ("name=%s\n",gtk_widget_get_name(widget));
		g_print ("value=%f\n",gtk_spin_button_get_value(GTK_SPIN_BUTTON (widget)));	
		g_print ("channel=%i send=%i\n",c,s);	
	}	
}

static void gtk_fader_gain_l_send_change(GtkWidget *widget, gpointer data)
{
	int i,h,c,s;
	char chan[10];
	char send[10];
	char name[10];
	double gain=phat_knob_get_value(PHAT_KNOB(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;
	
	sprintf(name,"%s",gtk_widget_get_name(widget));
	
	i=0;
	h=0;
	while (name[i] != ','){
		chan[i] = name[i];
		i++;
	}
	if (name[i] == ','){
		chan[i] = '\0';
		i++;
	}
	while (name[i] != '\0'){
		send[h] = name[i];
		i++;
		h++;
	}
	if (name[i] == '\0'){
		send[h] = '\0';
	}

	c=atoi(chan);
	s=atoi(send);
	
	if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_l)) != gain ){
		gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_l),gain);
	}
	
	if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_l_set", "iif",c,s,gain) == -1 ){
		printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
	}
					
	if (interface->verbose) {	
	    g_print ("Gtk send gain left fader changed\n");
		g_print ("name=%s\n",gtk_widget_get_name(widget));
		g_print ("value=%f\n",phat_knob_get_value(PHAT_KNOB(widget)));	
		g_print ("channel=%i send=%i\n",c,s);	
	}	
}

static void gtk_entry_gain_l_send_change(GtkWidget *widget, gpointer data)
{
	int i,h,c,s;
	char chan[10];
	char send[10];
	char name[10];
	double gain=gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;
	
	sprintf(name,"%s",gtk_widget_get_name(widget));
	
	i=0;
	h=0;
	while (name[i] != ','){
		chan[i] = name[i];
		i++;
	}
	if (name[i] == ','){
		chan[i] = '\0';
		i++;
	}
	while (name[i] != '\0'){
		send[h] = name[i];
		i++;
		h++;
	}
	if (name[i] == '\0'){
		send[h] = '\0';
	}

	c=atoi(chan);
	s=atoi(send);
	
	if (phat_knob_get_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_l)) != gain ){
		phat_knob_set_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_l),gain);
	}
			
	if (interface->verbose) {	
	    g_print ("Gtk send gain left entry changed\n");
		g_print ("name=%s\n",gtk_widget_get_name(widget));
		g_print ("value=%f\n",gtk_spin_button_get_value(GTK_SPIN_BUTTON (widget)));	
		g_print ("channel=%i send=%i\n",c,s);	
	}	
}

static void gtk_fader_gain_r_send_change(GtkWidget *widget, gpointer data)
{
	int i,h,c,s;
	char chan[10];
	char send[10];
	char name[10];
	double gain=phat_knob_get_value(PHAT_KNOB(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;
	
	sprintf(name,"%s",gtk_widget_get_name(widget));
	
	i=0;
	h=0;
	while (name[i] != ','){
		chan[i] = name[i];
		i++;
	}
	if (name[i] == ','){
		chan[i] = '\0';
		i++;
	}
	while (name[i] != '\0'){
		send[h] = name[i];
		i++;
		h++;
	}
	if (name[i] == '\0'){
		send[h] = '\0';
	}

	c=atoi(chan);
	s=atoi(send);
			
	if (gtk_spin_button_get_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_r)) != gain ){
		gtk_spin_button_set_value(GTK_SPIN_BUTTON (interface->channels[c-1].sends[s-1].entry_gain_r),gain);
	}
	
	if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE,"/mixer/channel/send/gain_r_set", "iif",c,s,gain) == -1 ){
		printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
	}
					
	if (interface->verbose) {	
	    g_print ("Gtk send gain right fader changed\n");
		g_print ("name=%s\n",gtk_widget_get_name(widget));
		g_print ("value=%f\n",phat_knob_get_value(PHAT_KNOB(widget)));	
		g_print ("channel=%i send=%i\n",c,s);	
	}	
}

static void gtk_entry_gain_r_send_change(GtkWidget *widget, gpointer data)
{
	int i,h,c,s;
	char chan[10];
	char send[10];
	char name[10];
	double gain=gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
	jmd_interface_t *interface = (jmd_interface_t*)data;
	
	sprintf(name,"%s",gtk_widget_get_name(widget));
	
	i=0;
	h=0;
	while (name[i] != ','){
		chan[i] = name[i];
		i++;
	}
	if (name[i] == ','){
		chan[i] = '\0';
		i++;
	}
	while (name[i] != '\0'){
		send[h] = name[i];
		i++;
		h++;
	}
	if (name[i] == '\0'){
		send[h] = '\0';
	}

	c=atoi(chan);
	s=atoi(send);
	
	if (phat_knob_get_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_r)) != gain ){
		phat_knob_set_value(PHAT_KNOB (interface->channels[c-1].sends[s-1].knob_gain_r),gain);
	}
			
	if (interface->verbose) {	
	    g_print ("Gtk send gain right entry changed\n");
		g_print ("name=%s\n",gtk_widget_get_name(widget));
		g_print ("value=%f\n",gtk_spin_button_get_value(GTK_SPIN_BUTTON (widget)));
		g_print ("channel=%i send=%i\n",c,s);	
	}	
}

static void gtk_about(GtkWidget *widget, gpointer data)
{
	GtkWidget *about_dialog;
	GtkWidget *about_button;
	GtkWidget *about_text_upper;
	GtkWidget *about_text_middle;
	GtkWidget *about_text_lower;
	GtkWidget *box_hbox;
	char window_title[10];
	
	box_hbox=gtk_hbox_new (TRUE, 10);
	about_text_upper =gtk_label_new("  Copyright (C) 2007  Ulrich Lorenz Schlueter  ");
	about_text_middle =gtk_label_new("audio-mobster@gmx.de");
	about_text_lower=gtk_label_new("GNU General Public License");
	about_dialog=gtk_dialog_new();
	about_button=gtk_button_new_with_label("Close");
	
	g_signal_connect_swapped (G_OBJECT (about_button), "clicked",
			      G_CALLBACK (gtk_widget_destroy),
                              G_OBJECT (about_dialog));
	
	sprintf(window_title,"About JackMixDesk_gtk");
	gtk_window_set_title (GTK_WINDOW (about_dialog), window_title);
	
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (about_dialog)->vbox),about_text_upper , TRUE, TRUE, 5);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (about_dialog)->vbox),about_text_middle , TRUE, TRUE, 5);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (about_dialog)->vbox),about_text_lower , TRUE, TRUE, 5);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (about_dialog)->action_area),about_button , TRUE, TRUE, 0);
                       
                
    gtk_widget_show (about_button);
    gtk_widget_show (about_text_upper);
    gtk_widget_show (about_text_middle);
    gtk_widget_show (about_text_lower);
    gtk_widget_show (box_hbox);
    gtk_widget_show (about_dialog);
}

static void gtk_destroy(GtkWidget *widget, gpointer data)
{
	running=0;
	g_print ("destroy event\n");
    gtk_main_quit ();
}

static void gtk_delete(GtkWidget *widget, gpointer data)
{
	running=0;
	g_print ("delete event\n");
    gtk_main_quit ();
}

static jmd_interface_t* interface_init()
{	
	jmd_interface_t *interface;
	interface = malloc(sizeof(jmd_interface_t));
	interface->verbose = 0;
	interface->quiet = 0;
	interface->st =NULL;
	interface->channel_count=0;
	interface->prefader_count=0;
	interface->postfader_count=0;
	interface->omit_send_pan=0;
	interface->omit_send_gain=0;
	interface->osc_mixer_port=DEFAULT_MIXER_PORT;
	interface->osc_mixer_host=DEFAULT_MIXER_HOST;
	
	snprintf( interface->client_name, DEFAULT_LABEL_LEN, "%s", DEFAULT_CLIENT_NAME );
	
	return(interface);
}

static int interface_usage()
{
	printf("Copyright (C) 2007 Ulrich Lorenz Schlueter <audio-mobster@gmx.de>\n\n");
	printf("This program comes with ABSOLUTELY NO WARRANTY.  You are licensed to use it\n");
	printf("under the terms of the GNU General Public License, version 2 or later.  See\n");
	printf("the COPYING file that came with this software for details.\n\n");
	printf("JackMixDesk_gtk\n\n");
	printf("   -p <port>     Set the UDP port number for jackmixdesk.\n");
	printf("   -m <host>     Set the UDP hostname for jackmixdesk (default localhost).\n");
	printf("   -n <name>     Name for this OSC client (default jackmixdesk_gtk).\n");
	printf("   -d            Omit send pan knobs use send gain left and right instead.\n");
	printf("   -g            Omit send send gain left and right knobs use pan instead.\n");
	printf("   -c            Omit main channel pan use....\n");
	printf("   -v            Enable verbose mode.\n");
	printf("   -q            Enable quiet mode.\n");
	printf("\n");
	exit(1);
}

static int interface_init_mixer(jmd_interface_t *interface){
	
	interface->t = lo_address_new(interface->osc_mixer_host, interface->osc_mixer_port);
	interface->serv = lo_server_thread_get_server( interface->st );

	if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel_count_get", "") == -1) {
		printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
	}
	if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/prefader_count_get", "") == -1) {
		printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
	}
	if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/postfader_count_get", "") == -1) {
		printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
	}

	return 0; 
}

static int interface_refresh_channels(jmd_interface_t *interface)
{
	int i,h;
		
	for (i=0;i<interface->channel_count;i++){
		
	    if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/label_get", "i",i+1) == -1) {
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
		if(i-1 < interface->channel_count){
	    	if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/stereo_get", "i",i+1) == -1) {
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
		}
	    if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/mute_get", "i",i+1) == -1) {
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
		
		for (h=0;h<interface->prefader_count+interface->postfader_count;h++){
			
			if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/send/gain_get", "ii",i+1,h+1) == -1) {
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
			if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/send/pan_get", "ii",i+1,h+1) == -1) {
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
			if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/send/gain_l_get", "ii",i+1,h+1) == -1) {
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
			if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/send/gain_r_get", "ii",i+1,h+1) == -1) {
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
		}
	}
	return 0;
}

static int interface_init_channels(jmd_interface_t *interface)
{
	int i,h;
	
	interface->channels = malloc( 
		sizeof(jmd_channel_t) * interface->channel_count);
		
	for (i=0;i<interface->channel_count;i++){
		
		interface->channels[i].sends = malloc( 
			sizeof(jmd_send_t) * (interface->prefader_count+interface->postfader_count));
		
	    if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/label_get", "i",i+1) == -1) {
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
		if(i-1 < interface->channel_count){
	    	if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/stereo_get", "i",i+1) == -1) {
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
		}
	    if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/mute_get", "i",i+1) == -1) {
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
		
		for (h=0;h<interface->prefader_count+interface->postfader_count;h++){
			
			if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/send/gain_get", "ii",i+1,h+1) == -1) {
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
			if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/send/pan_get", "ii",i+1,h+1) == -1) {
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
			if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/send/gain_l_get", "ii",i+1,h+1) == -1) {
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
			if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/channel/send/gain_r_get", "ii",i+1,h+1) == -1) {
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
		}
	}
	return 0;
}

static int interface_init_outputs_struct(jmd_interface_t *interface)
{
	interface->outputs = malloc( 
		sizeof(jmd_output_t) * (interface->prefader_count+interface->postfader_count));

	return 0;
}

static int interface_init_outputs_data(jmd_interface_t *interface)
{
	int i;

	for (i=0;i<interface->prefader_count+interface->postfader_count;i++){
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/output/label_get", "i",i+1) == -1) {
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/output/gain_get", "i",i+1) == -1) {
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/output/mute_get", "i",i+1) == -1) {
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
		if(i < interface->prefader_count+interface->postfader_count-1){
	    	if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/output/link_get", "i",i+1) == -1) {
				printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
			}
		}
	}
	return 0;
}

static int interface_init_gtk_channels_1(jmd_interface_t *interface)
{
	char window_title[32];
	char lc_name[5];
	char *utf_name=NULL;
	int c,s;
	double pan;
	char chan_label[10];
	
    interface->window_channel_1 = gtk_window_new(GTK_WINDOW_TOPLEVEL);

    interface->box_channels_1_hbox = gtk_hbox_new (FALSE, 0);
    
    gtk_widget_show (interface->box_channels_1_hbox);
    
	interface->window_channels_1 = gtk_scrolled_window_new(NULL, NULL);
    
	sprintf( window_title,"JackMixDesk Channels 1 - %i", interface->channel_count);
	gtk_window_set_title (GTK_WINDOW (interface->window_channel_1), window_title);
	
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (interface->window_channels_1),
                                    GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
                                    
    gtk_container_add (GTK_CONTAINER (interface->window_channel_1),interface->window_channels_1 );                            

	gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (interface->window_channels_1), interface->box_channels_1_hbox);

	g_signal_connect_swapped (G_OBJECT (interface->window_channel_1), "delete_event",
			      G_CALLBACK (gtk_widget_destroy),
                              G_OBJECT (interface->window_channel_1));

	g_signal_connect_swapped (G_OBJECT (interface->window_channel_1), "destroy",
			      G_CALLBACK (gtk_widget_destroy),
                              G_OBJECT (interface->window_channel_1));

    gtk_container_set_border_width (GTK_CONTAINER (interface->window_channel_1), 0);

    gtk_widget_set_size_request (interface->window_channel_1, 768, 580);

    /* Sets the border width of the window. */
    gtk_container_set_border_width (GTK_CONTAINER (interface->window_channels_1), 0);
    
    /* Style for spin buttons */
    interface->spinbutton_style = gtk_style_new();
    interface->small_label_style = gtk_style_new();
    interface->medium_label_style = gtk_style_new();

    /* set fonts */
    pango_font_description_set_absolute_size(interface->spinbutton_style->font_desc,SPINBUTTON_FONT_SIZE);
    pango_font_description_set_absolute_size(interface->small_label_style->font_desc,SMALL_LABEL_FONT_SIZE);
    pango_font_description_set_absolute_size(interface->medium_label_style->font_desc,MEDIUM_LABEL_FONT_SIZE);
	pango_font_description_set_family(interface->spinbutton_style->font_desc,"bold");
	pango_font_description_set_family(interface->small_label_style->font_desc,"bold");
	pango_font_description_set_family(interface->medium_label_style->font_desc,"bold");
	
	/* create the widgets for the channel window */
    for(c=0;c<interface->channel_count;c++){
    	
		if (interface->channels[c].sends[0].gain_l == 0 && interface->channels[c].sends[0].gain_r < 0){
			pan = interface->channels[c].sends[0].gain_r;
		} else if (interface->channels[c].sends[0].gain_r == 0 && interface->channels[c].sends[0].gain_l < 0){
			pan = interface->channels[c].sends[0].gain_l * (-1);
		} else {
			pan = 0;
		}
    	
    	sprintf(chan_label,"%s",interface->channels[c].label);
    	interface->channels[c].box_vbox = gtk_vbox_new (FALSE, 0);
    	interface->channels[c].box_hbox_button_stereo = gtk_hbox_new (FALSE, 0);
		gtk_widget_set_size_request(interface->channels[c].box_hbox_button_stereo,-1,20);

    	interface->channels[c].box_hbox_button_mute = gtk_hbox_new (FALSE, 0);
    	interface->channels[c].box_hbox_entry_gain = gtk_hbox_new (FALSE, 0);
    	interface->channels[c].label_gain = gtk_label_new("Gain");
    	interface->channels[c].fader_gain = phat_vfan_slider_new_with_range(interface->channels[c].sends[0].gain,KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);		
		/* omit for last channel */
		if( c < interface->channel_count -1){
    		interface->channels[c].label_stereo = gtk_label_new("Stereo");
     		interface->channels[c].button_stereo = gtk_check_button_new();
		}
     	interface->channels[c].button_mute = gtk_check_button_new_with_label("Mute");

		interface->channels[c].entry_gain = gtk_spin_button_new_with_range(KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);
		gtk_widget_set_style(interface->channels[c].entry_gain,interface->spinbutton_style);
    	interface->channels[c].label_channel_top = gtk_label_new(chan_label);
		interface->channels[c].separator_top = gtk_hseparator_new ();
		interface->channels[c].separator_bottom = gtk_hseparator_new ();
    	interface->channels[c].label_channel_bottom = gtk_label_new(chan_label);
		gtk_entry_set_width_chars(GTK_ENTRY(interface->channels[c].entry_gain),DEFAULT_ENTRY_LEN);
		gtk_spin_button_set_digits(GTK_SPIN_BUTTON(interface->channels[c].entry_gain),1);
		sprintf(lc_name,"%i",c+1);
		utf_name = stringprep_locale_to_utf8(lc_name);
		gtk_widget_set_name (interface->channels[c].fader_gain,utf_name);
		gtk_widget_set_name (interface->channels[c].entry_gain,utf_name);
		/* omit for last channel */
		if( c < interface->channel_count -1){
    		gtk_widget_set_name (interface->channels[c].button_stereo,utf_name);
		}
		gtk_widget_set_name (interface->channels[c].button_mute,utf_name);
		
		g_signal_connect (G_OBJECT (interface->channels[c].fader_gain), "value_changed",
				G_CALLBACK (gtk_fader_gain_chan_change),(gpointer) interface );

		g_signal_connect (G_OBJECT (interface->channels[c].entry_gain), "value_changed",
				G_CALLBACK (gtk_entry_gain_chan_change),(gpointer) interface );

		/* omit for last channel */
		if( c < interface->channel_count -1){
    			g_signal_connect (G_OBJECT (interface->channels[c].button_stereo), "clicked",
					G_CALLBACK (gtk_button_stereo_chan_change),(gpointer) interface );
		}

		g_signal_connect (G_OBJECT (interface->channels[c].button_mute), "clicked",
				G_CALLBACK (gtk_button_mute_chan_change),(gpointer) interface );

    	gtk_widget_show (interface->channels[c].box_vbox);
    	gtk_widget_show (interface->channels[c].box_hbox_entry_gain);
    	gtk_widget_show (interface->channels[c].box_hbox_button_stereo);
    	gtk_widget_show (interface->channels[c].box_hbox_button_mute);
    	gtk_widget_show (interface->channels[c].label_gain);
    	gtk_widget_show (interface->channels[c].fader_gain);
    	gtk_widget_show (interface->channels[c].entry_gain);
		/* omit for last channel */
		if( c < interface->channel_count -1){
        	gtk_widget_show (interface->channels[c].label_stereo);
    		gtk_widget_show (interface->channels[c].button_stereo);
		}
    	gtk_widget_show (interface->channels[c].button_mute);
    	gtk_widget_show (interface->channels[c].label_channel_top);
    	gtk_widget_show (interface->channels[c].label_channel_bottom);
    	gtk_widget_show (interface->channels[c].separator_top);
    	gtk_widget_show (interface->channels[c].separator_bottom);
    	
    	gtk_box_pack_start (GTK_BOX (interface->box_channels_1_hbox), interface->channels[c].box_vbox, FALSE, FALSE, FALSE);
    	gtk_box_pack_start (GTK_BOX ( interface->channels[c].box_vbox), interface->channels[c].label_channel_top, FALSE, FALSE, FALSE);
    	gtk_box_pack_start (GTK_BOX ( interface->channels[c].box_vbox), interface->channels[c].separator_top, FALSE, FALSE, FALSE);
    	gtk_box_pack_end (GTK_BOX ( interface->channels[c].box_vbox), interface->channels[c].label_channel_bottom, FALSE, FALSE, FALSE);
    	gtk_box_pack_end (GTK_BOX ( interface->channels[c].box_vbox), interface->channels[c].separator_bottom, FALSE, FALSE, FALSE);
    	gtk_box_pack_end (GTK_BOX ( interface->channels[c].box_vbox), interface->channels[c].box_hbox_button_stereo, FALSE, FALSE, FALSE);
    	gtk_box_pack_end (GTK_BOX ( interface->channels[c].box_vbox), interface->channels[c].box_hbox_button_mute, FALSE, FALSE, FALSE);
    	gtk_box_pack_start (GTK_BOX ( interface->channels[c].box_hbox_button_mute), interface->channels[c].button_mute, FALSE, FALSE, FALSE);
		/* omit for last channel */
		if( c < interface->channel_count -1){
    		gtk_box_pack_end (GTK_BOX ( interface->channels[c].box_hbox_button_stereo), interface->channels[c].button_stereo, FALSE, FALSE, FALSE);
    		gtk_box_pack_end (GTK_BOX ( interface->channels[c].box_hbox_button_stereo), interface->channels[c].label_stereo, FALSE, FALSE, FALSE);
		}
    	gtk_box_pack_end (GTK_BOX ( interface->channels[c].box_vbox), interface->channels[c].box_hbox_entry_gain, FALSE, TRUE, FALSE);
    	gtk_box_pack_end (GTK_BOX ( interface->channels[c].box_hbox_entry_gain), interface->channels[c].entry_gain, TRUE, FALSE, TRUE);
    	gtk_box_pack_end (GTK_BOX ( interface->channels[c].box_vbox), interface->channels[c].fader_gain, TRUE, TRUE, TRUE);
    	gtk_box_pack_end (GTK_BOX ( interface->channels[c].box_vbox), interface->channels[c].label_gain, FALSE, FALSE, FALSE);

    	if (!interface->omit_chan_pan){
    		
    		interface->channels[c].box_hbox_knob_pan = gtk_hbox_new (FALSE, 0);
    		interface->channels[c].box_hbox_entry_pan = gtk_hbox_new (FALSE, 0);
    		interface->channels[c].label_pan = gtk_label_new("Pan");    		
	    	interface->channels[c].knob_pan = phat_knob_new_with_range(pan,KNOB_PAN_MIN,KNOB_PAN_MAX,KNOB_PAN_STEP);
	    	interface->channels[c].entry_pan = gtk_spin_button_new_with_range(KNOB_PAN_MIN,KNOB_PAN_MAX,KNOB_PAN_STEP);

			gtk_widget_set_name (interface->channels[c].knob_pan,utf_name);
			gtk_widget_set_name (interface->channels[c].entry_pan,utf_name);

			g_signal_connect (G_OBJECT (interface->channels[c].knob_pan), "value_changed",
					G_CALLBACK (gtk_fader_pan_chan_change),(gpointer) interface );

			g_signal_connect (G_OBJECT (interface->channels[c].entry_pan), "value_changed",
					G_CALLBACK (gtk_entry_pan_chan_change),(gpointer) interface );

			gtk_widget_set_size_request (interface->channels[c].knob_pan, 50, 50);
			gtk_widget_set_style(interface->channels[c].entry_pan,interface->spinbutton_style);
			gtk_spin_button_set_digits(GTK_SPIN_BUTTON(interface->channels[c].entry_pan),1);
	    	gtk_entry_set_width_chars(GTK_ENTRY(interface->channels[c].entry_pan),DEFAULT_ENTRY_LEN);
	    	
	    	gtk_widget_show (interface->channels[c].box_hbox_knob_pan);
	    	gtk_widget_show (interface->channels[c].box_hbox_entry_pan);
	    	gtk_widget_show (interface->channels[c].label_pan);
	    	gtk_widget_show (interface->channels[c].knob_pan);
	    	gtk_widget_show (interface->channels[c].entry_pan);
	    	
	    	gtk_box_pack_end (GTK_BOX ( interface->channels[c].box_vbox), interface->channels[c].box_hbox_entry_pan, FALSE, TRUE, FALSE);
	    	gtk_box_pack_start (GTK_BOX ( interface->channels[c].box_hbox_entry_pan), interface->channels[c].entry_pan, TRUE, FALSE, TRUE);
	    	gtk_box_pack_end (GTK_BOX ( interface->channels[c].box_vbox), interface->channels[c].box_hbox_knob_pan, FALSE, FALSE, FALSE);
	    	gtk_box_pack_end (GTK_BOX ( interface->channels[c].box_vbox), interface->channels[c].label_pan, FALSE, FALSE, FALSE);
	    	gtk_box_pack_start (GTK_BOX ( interface->channels[c].box_hbox_knob_pan), interface->channels[c].knob_pan, TRUE, FALSE, TRUE);
    	}

		for(s=0;s<(interface->prefader_count+interface->postfader_count);s++){
			
			interface->channels[c].sends[s].label_send = gtk_label_new(interface->channels[c].sends[s].label);
			interface->channels[c].sends[s].box_vbox = gtk_vbox_new (FALSE, 0);
			interface->channels[c].sends[s].box_hbox_upper = gtk_hbox_new (FALSE, 0);
			interface->channels[c].sends[s].box_hbox_middle = gtk_hbox_new (FALSE, 0);
			interface->channels[c].sends[s].box_hbox_lower = gtk_hbox_new (FALSE, 0);
			interface->channels[c].sends[s].separator = gtk_hseparator_new ();
			interface->channels[c].sends[s].label_gain = gtk_label_new("Gain");
			
			gtk_box_set_spacing(GTK_BOX(interface->channels[c].box_vbox),3);
			gtk_widget_set_style(interface->channels[c].sends[s].label_send,interface->medium_label_style);
			gtk_widget_set_style(interface->channels[c].sends[s].label_gain,interface->small_label_style);
			
			interface->channels[c].sends[s].knob_gain = phat_knob_new_with_range(interface->channels[c].sends[s].gain,KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);
			interface->channels[c].sends[s].entry_gain = gtk_spin_button_new_with_range(KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);

			sprintf(lc_name,"%i,%i",c+1,s+1);
			utf_name = stringprep_locale_to_utf8(lc_name);
			gtk_widget_set_name (interface->channels[c].sends[s].knob_gain,utf_name);
			gtk_widget_set_name (interface->channels[c].sends[s].entry_gain,utf_name);

			gtk_widget_set_size_request (interface->channels[c].sends[s].knob_gain, KNOB_SIZE, KNOB_SIZE);
			gtk_widget_set_style(interface->channels[c].sends[s].entry_gain,interface->spinbutton_style);
			gtk_entry_set_width_chars(GTK_ENTRY(interface->channels[c].sends[s].entry_gain),DEFAULT_ENTRY_LEN);
			gtk_spin_button_set_digits(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_gain),1);

			g_signal_connect (G_OBJECT (interface->channels[c].sends[s].knob_gain), "value_changed",
								G_CALLBACK (gtk_fader_gain_send_change),(gpointer) interface);
			
			g_signal_connect (G_OBJECT (interface->channels[c].sends[s].entry_gain), "value_changed",
								G_CALLBACK (gtk_entry_gain_send_change),(gpointer) interface);
			
			gtk_box_pack_start (GTK_BOX (interface->channels[c].box_vbox),interface->channels[c].sends[s].box_vbox, FALSE, FALSE, FALSE);
			gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox),interface->channels[c].sends[s].box_hbox_upper, FALSE, FALSE, FALSE);
			gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_upper),interface->channels[c].sends[s].label_send, TRUE, FALSE, FALSE);
			
			gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox),interface->channels[c].sends[s].box_hbox_middle, FALSE, FALSE, FALSE);
		    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox),interface->channels[c].sends[s].box_hbox_lower, FALSE, FALSE, FALSE);
		    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox), interface->channels[c].sends[s].separator, FALSE, FALSE, FALSE);
			
			gtk_widget_show (interface->channels[c].sends[s].label_send);
			gtk_widget_show (interface->channels[c].sends[s].box_vbox);
			gtk_widget_show (interface->channels[c].sends[s].box_hbox_upper);
			gtk_widget_show (interface->channels[c].sends[s].box_hbox_middle);
			gtk_widget_show (interface->channels[c].sends[s].box_hbox_lower);
			gtk_widget_show (interface->channels[c].sends[s].separator);
			
			if (!interface->omit_send_pan && !interface->omit_send_gain){
				
				interface->channels[c].sends[s].box_vbox_middle_left = gtk_vbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_vbox_middle_right = gtk_vbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_middle_knob_left = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_middle_knob_right = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_middle_entry_left = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_middle_entry_right = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_vbox_lower_left = gtk_vbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_vbox_lower_right = gtk_vbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_lower_knob_left = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_lower_knob_right = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_lower_entry_left = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_lower_entry_right = gtk_hbox_new (FALSE, 0);
				
				interface->channels[c].sends[s].label_pan = gtk_label_new("Pan");
				interface->channels[c].sends[s].label_gain_l = gtk_label_new("Gain L");
				interface->channels[c].sends[s].label_gain_r = gtk_label_new("Gain R");	
				
				gtk_widget_set_style(interface->channels[c].sends[s].label_pan,interface->small_label_style);
				gtk_widget_set_style(interface->channels[c].sends[s].label_gain_l,interface->small_label_style);
				gtk_widget_set_style(interface->channels[c].sends[s].label_gain_r,interface->small_label_style);

				interface->channels[c].sends[s].knob_pan = phat_knob_new_with_range(pan,KNOB_PAN_MIN,KNOB_PAN_MAX,KNOB_PAN_STEP);
				interface->channels[c].sends[s].knob_gain_l = phat_knob_new_with_range(interface->channels[c].sends[s].gain_l,KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);
				interface->channels[c].sends[s].knob_gain_r = phat_knob_new_with_range(interface->channels[c].sends[s].gain_r,KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);
				
				gtk_widget_set_size_request (interface->channels[c].sends[s].knob_pan, KNOB_SIZE, KNOB_SIZE);
				gtk_widget_set_size_request (interface->channels[c].sends[s].knob_gain_l, KNOB_SIZE, KNOB_SIZE);
				gtk_widget_set_size_request (interface->channels[c].sends[s].knob_gain_r, KNOB_SIZE, KNOB_SIZE);

				interface->channels[c].sends[s].entry_pan = gtk_spin_button_new_with_range(KNOB_PAN_MIN,KNOB_PAN_MAX,KNOB_PAN_STEP);
				interface->channels[c].sends[s].entry_gain_l = gtk_spin_button_new_with_range(KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);
				interface->channels[c].sends[s].entry_gain_r = gtk_spin_button_new_with_range(KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);

				gtk_spin_button_set_digits(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_pan),1);
				gtk_spin_button_set_digits(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_gain_l),1);
				gtk_spin_button_set_digits(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_gain_r),1);

				gtk_widget_set_style(interface->channels[c].sends[s].entry_pan,interface->spinbutton_style);
				gtk_widget_set_style(interface->channels[c].sends[s].entry_gain_l,interface->spinbutton_style);
				gtk_widget_set_style(interface->channels[c].sends[s].entry_gain_r,interface->spinbutton_style);
	
				gtk_widget_set_name (interface->channels[c].sends[s].knob_pan,utf_name);
				gtk_widget_set_name (interface->channels[c].sends[s].entry_pan,utf_name);
				gtk_widget_set_name (interface->channels[c].sends[s].knob_gain_l,utf_name);
				gtk_widget_set_name (interface->channels[c].sends[s].entry_gain_l,utf_name);
				gtk_widget_set_name (interface->channels[c].sends[s].knob_gain_r,utf_name);
				gtk_widget_set_name (interface->channels[c].sends[s].entry_gain_r,utf_name);

				gtk_entry_set_width_chars(GTK_ENTRY(interface->channels[c].sends[s].entry_pan),DEFAULT_ENTRY_LEN);
				gtk_entry_set_width_chars(GTK_ENTRY(interface->channels[c].sends[s].entry_gain_l),DEFAULT_ENTRY_LEN);
				gtk_entry_set_width_chars(GTK_ENTRY(interface->channels[c].sends[s].entry_gain_r),DEFAULT_ENTRY_LEN);
				
			    g_signal_connect (G_OBJECT (interface->channels[c].sends[s].knob_pan), "value_changed",
					      G_CALLBACK (gtk_fader_pan_send_change),(gpointer) interface);
			    
			    g_signal_connect (G_OBJECT (interface->channels[c].sends[s].knob_gain_l), "value_changed",
				      G_CALLBACK (gtk_fader_gain_l_send_change),(gpointer) interface);
		                              
			    g_signal_connect (G_OBJECT (interface->channels[c].sends[s].knob_gain_r), "value_changed",
				      G_CALLBACK (gtk_fader_gain_r_send_change),(gpointer) interface);

			    g_signal_connect (G_OBJECT (interface->channels[c].sends[s].entry_pan), "value_changed",
					      G_CALLBACK (gtk_entry_pan_send_change),(gpointer) interface);
			    
			    g_signal_connect (G_OBJECT (interface->channels[c].sends[s].entry_gain_l), "value_changed",
				      G_CALLBACK (gtk_entry_gain_l_send_change),(gpointer) interface);
		                              
			    g_signal_connect (G_OBJECT (interface->channels[c].sends[s].entry_gain_r), "value_changed",
				      G_CALLBACK (gtk_entry_gain_r_send_change),(gpointer) interface);

			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle),interface->channels[c].sends[s].box_vbox_middle_left, FALSE, FALSE, FALSE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle),interface->channels[c].sends[s].box_vbox_middle_right, FALSE, FALSE, FALSE);
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_lower),interface->channels[c].sends[s].box_vbox_lower_left, FALSE, FALSE, FALSE);
		    	gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_lower),interface->channels[c].sends[s].box_vbox_lower_right, FALSE, FALSE, FALSE);			                              

			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_left),interface->channels[c].sends[s].label_gain, FALSE, FALSE, FALSE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_right),interface->channels[c].sends[s].label_pan, FALSE, FALSE, FALSE);
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_lower_left),interface->channels[c].sends[s].label_gain_l, FALSE, FALSE, FALSE);
		    	gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_lower_right),interface->channels[c].sends[s].label_gain_r, FALSE, FALSE, FALSE);			                              
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_left),interface->channels[c].sends[s].box_hbox_middle_knob_left, FALSE, TRUE, FALSE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_right),interface->channels[c].sends[s].box_hbox_middle_knob_right, FALSE, TRUE, FALSE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle_knob_left),interface->channels[c].sends[s].knob_gain, TRUE, FALSE, TRUE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle_knob_right),interface->channels[c].sends[s].knob_pan, TRUE, FALSE, TRUE);
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_lower_left),interface->channels[c].sends[s].box_hbox_lower_knob_left,  FALSE, TRUE, FALSE);
		    	gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_lower_right),interface->channels[c].sends[s].box_hbox_lower_knob_right,  FALSE, TRUE, FALSE);			                              
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_lower_knob_left),interface->channels[c].sends[s].knob_gain_l, TRUE, FALSE, TRUE);
		    	gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_lower_knob_right),interface->channels[c].sends[s].knob_gain_r, TRUE, FALSE, TRUE);			                              
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_left),interface->channels[c].sends[s].box_hbox_middle_entry_left, FALSE, TRUE, FALSE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_right),interface->channels[c].sends[s].box_hbox_middle_entry_right, FALSE, TRUE, FALSE);
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_lower_left),interface->channels[c].sends[s].box_hbox_lower_entry_left, FALSE, TRUE, FALSE);
		    	gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_lower_right),interface->channels[c].sends[s].box_hbox_lower_entry_right, FALSE, TRUE, FALSE);			                              
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle_entry_left),interface->channels[c].sends[s].entry_gain, TRUE, FALSE, TRUE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle_entry_right),interface->channels[c].sends[s].entry_pan, TRUE, FALSE, TRUE);
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_lower_entry_left),interface->channels[c].sends[s].entry_gain_l, TRUE, FALSE, TRUE);
		    	gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_lower_entry_right),interface->channels[c].sends[s].entry_gain_r, TRUE, FALSE, TRUE);			                              

				gtk_widget_show (interface->channels[c].sends[s].box_vbox_middle_left);
				gtk_widget_show (interface->channels[c].sends[s].box_vbox_middle_right);
				gtk_widget_show (interface->channels[c].sends[s].box_vbox_lower_left);
				gtk_widget_show (interface->channels[c].sends[s].box_vbox_lower_right);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_middle_knob_left);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_middle_knob_right);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_middle_entry_left);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_middle_entry_right);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_lower_knob_left);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_lower_knob_right);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_lower_entry_left);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_lower_entry_right);
				
			    gtk_widget_show (interface->channels[c].sends[s].label_pan);
		    	gtk_widget_show (interface->channels[c].sends[s].label_gain_l);
		    	gtk_widget_show (interface->channels[c].sends[s].label_gain_r);
			    gtk_widget_show (interface->channels[c].sends[s].knob_pan);
		    	gtk_widget_show (interface->channels[c].sends[s].knob_gain_l);
		    	gtk_widget_show (interface->channels[c].sends[s].knob_gain_r);
			    gtk_widget_show (interface->channels[c].sends[s].entry_pan);
		    	gtk_widget_show (interface->channels[c].sends[s].entry_gain_l);
		    	gtk_widget_show (interface->channels[c].sends[s].entry_gain_r);
		    	
			}
			
			if (interface->omit_send_pan && !interface->omit_send_gain){
				
				interface->channels[c].sends[s].box_vbox_middle_left = gtk_vbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_middle_knob_left = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_middle_entry_left = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_vbox_lower_left = gtk_vbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_vbox_lower_right = gtk_vbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_lower_knob_left = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_lower_knob_right = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_lower_entry_left = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_lower_entry_right = gtk_hbox_new (FALSE, 0);
				
				interface->channels[c].sends[s].label_gain_l = gtk_label_new("Gain L");
				interface->channels[c].sends[s].label_gain_r = gtk_label_new("Gain R");

				gtk_widget_set_style(interface->channels[c].sends[s].label_gain_l,interface->small_label_style);
				gtk_widget_set_style(interface->channels[c].sends[s].label_gain_r,interface->small_label_style);


				interface->channels[c].sends[s].knob_gain_l = phat_knob_new_with_range(interface->channels[c].sends[s].gain_l,KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);
				interface->channels[c].sends[s].knob_gain_r = phat_knob_new_with_range(interface->channels[c].sends[s].gain_r,KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);
				interface->channels[c].sends[s].entry_gain_l = gtk_spin_button_new_with_range(KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);
				interface->channels[c].sends[s].entry_gain_r = gtk_spin_button_new_with_range(KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);


				gtk_widget_set_name (interface->channels[c].sends[s].knob_gain_l,utf_name);
				gtk_widget_set_name (interface->channels[c].sends[s].entry_gain_l,utf_name);
				gtk_widget_set_name (interface->channels[c].sends[s].knob_gain_r,utf_name);
				gtk_widget_set_name (interface->channels[c].sends[s].entry_gain_r,utf_name);

				gtk_widget_set_size_request (interface->channels[c].sends[s].knob_gain_l, KNOB_SIZE, KNOB_SIZE);
				gtk_widget_set_size_request (interface->channels[c].sends[s].knob_gain_r, KNOB_SIZE, KNOB_SIZE);
				gtk_entry_set_width_chars(GTK_ENTRY(interface->channels[c].sends[s].entry_gain_l),DEFAULT_ENTRY_LEN);
				gtk_entry_set_width_chars(GTK_ENTRY(interface->channels[c].sends[s].entry_gain_r),DEFAULT_ENTRY_LEN);
				gtk_widget_set_style(interface->channels[c].sends[s].entry_gain_l,interface->spinbutton_style);
				gtk_widget_set_style(interface->channels[c].sends[s].entry_gain_r,interface->spinbutton_style);
				gtk_spin_button_set_digits(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_gain_l),1);
				gtk_spin_button_set_digits(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_gain_r),1);
				
			    g_signal_connect (G_OBJECT (interface->channels[c].sends[s].knob_gain_l), "value_changed",
				      G_CALLBACK (gtk_fader_gain_l_send_change),(gpointer) interface);
		                              
			    g_signal_connect (G_OBJECT (interface->channels[c].sends[s].knob_gain_r), "value_changed",
				      G_CALLBACK (gtk_fader_gain_r_send_change),(gpointer) interface);

			    g_signal_connect (G_OBJECT (interface->channels[c].sends[s].entry_gain_l), "value_changed",
				      G_CALLBACK (gtk_entry_gain_l_send_change),(gpointer) interface);
		                              
			    g_signal_connect (G_OBJECT (interface->channels[c].sends[s].entry_gain_r), "value_changed",
				      G_CALLBACK (gtk_entry_gain_r_send_change),(gpointer) interface);

			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle),interface->channels[c].sends[s].box_vbox_middle_left, TRUE, FALSE, TRUE);
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_lower),interface->channels[c].sends[s].box_vbox_lower_left, FALSE, FALSE, FALSE);
		    	gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_lower),interface->channels[c].sends[s].box_vbox_lower_right, FALSE, FALSE, FALSE);			                              

			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_left),interface->channels[c].sends[s].label_gain, FALSE, FALSE, FALSE);
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_lower_left),interface->channels[c].sends[s].label_gain_l, FALSE, FALSE, FALSE);
		    	gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_lower_right),interface->channels[c].sends[s].label_gain_r, FALSE, FALSE, FALSE);			                              
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_left),interface->channels[c].sends[s].box_hbox_middle_knob_left, FALSE, TRUE, FALSE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle_knob_left),interface->channels[c].sends[s].knob_gain, TRUE, FALSE, TRUE);
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_lower_left),interface->channels[c].sends[s].box_hbox_lower_knob_left,  FALSE, TRUE, FALSE);
		    	gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_lower_right),interface->channels[c].sends[s].box_hbox_lower_knob_right,  FALSE, TRUE, FALSE);			                              
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_lower_knob_left),interface->channels[c].sends[s].knob_gain_l, TRUE, FALSE, TRUE);
		    	gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_lower_knob_right),interface->channels[c].sends[s].knob_gain_r, TRUE, FALSE, TRUE);			                              
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_left),interface->channels[c].sends[s].box_hbox_middle_entry_left, FALSE, TRUE, FALSE);
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_lower_left),interface->channels[c].sends[s].box_hbox_lower_entry_left, FALSE, TRUE, FALSE);
		    	gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_lower_right),interface->channels[c].sends[s].box_hbox_lower_entry_right, FALSE, TRUE, FALSE);			                              
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle_entry_left),interface->channels[c].sends[s].entry_gain, TRUE, FALSE, TRUE);
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_lower_entry_left),interface->channels[c].sends[s].entry_gain_l, TRUE, FALSE, TRUE);
		    	gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_lower_entry_right),interface->channels[c].sends[s].entry_gain_r, TRUE, FALSE, TRUE);			                              

				gtk_widget_show (interface->channels[c].sends[s].box_vbox_middle_left);
				gtk_widget_show (interface->channels[c].sends[s].box_vbox_lower_left);
				gtk_widget_show (interface->channels[c].sends[s].box_vbox_lower_right);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_middle_knob_left);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_middle_entry_left);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_lower_knob_left);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_lower_knob_right);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_lower_entry_left);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_lower_entry_right);
				
		    	gtk_widget_show (interface->channels[c].sends[s].label_gain_l);
		    	gtk_widget_show (interface->channels[c].sends[s].label_gain_r);
		    	gtk_widget_show (interface->channels[c].sends[s].knob_gain_l);
		    	gtk_widget_show (interface->channels[c].sends[s].knob_gain_r);
		    	gtk_widget_show (interface->channels[c].sends[s].entry_gain_l);
		    	gtk_widget_show (interface->channels[c].sends[s].entry_gain_r);

			}
			
			if (!interface->omit_send_pan && interface->omit_send_gain){
				
				interface->channels[c].sends[s].box_vbox_middle_left = gtk_vbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_vbox_middle_right = gtk_vbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_middle_knob_left = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_middle_knob_right = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_middle_entry_left = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_middle_entry_right = gtk_hbox_new (FALSE, 0);
				
				interface->channels[c].sends[s].label_pan = gtk_label_new("Pan");
				gtk_widget_set_style(interface->channels[c].sends[s].label_pan,interface->small_label_style);
				
				
				if (interface->channels[c].sends[s].gain_l == 0 && interface->channels[c].sends[s].gain_r < 0){
					pan = interface->channels[c].sends[s].gain_r;
				} else if (interface->channels[c].sends[s].gain_r == 0 && interface->channels[c].sends[s].gain_l < 0){
					pan = interface->channels[c].sends[s].gain_l * (-1);
				} else {
					pan = 0;
				}
					
				interface->channels[c].sends[s].knob_pan = phat_knob_new_with_range(pan,KNOB_PAN_MIN,KNOB_PAN_MAX,KNOB_PAN_STEP);
				interface->channels[c].sends[s].entry_pan = gtk_spin_button_new_with_range(KNOB_PAN_MIN,KNOB_PAN_MAX,KNOB_PAN_STEP);

				gtk_widget_set_style(interface->channels[c].sends[s].entry_pan,interface->spinbutton_style);
				gtk_spin_button_set_digits(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_pan),1);
				gtk_widget_set_size_request (interface->channels[c].sends[s].knob_pan, KNOB_SIZE, KNOB_SIZE);
				gtk_widget_set_name (interface->channels[c].sends[s].knob_pan,utf_name);
				gtk_widget_set_name (interface->channels[c].sends[s].entry_pan,utf_name);


				gtk_entry_set_width_chars(GTK_ENTRY(interface->channels[c].sends[s].entry_pan),DEFAULT_ENTRY_LEN);
				
			    g_signal_connect (G_OBJECT (interface->channels[c].sends[s].knob_pan), "value_changed",
					      G_CALLBACK (gtk_fader_pan_send_change),(gpointer) interface);

			    g_signal_connect (G_OBJECT (interface->channels[c].sends[s].entry_pan), "value_changed",
					      G_CALLBACK (gtk_entry_pan_send_change),(gpointer) interface);

			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle),interface->channels[c].sends[s].box_vbox_middle_left, FALSE, FALSE, FALSE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle),interface->channels[c].sends[s].box_vbox_middle_right, FALSE, FALSE, FALSE);			                              

			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_left),interface->channels[c].sends[s].label_gain, FALSE, FALSE, FALSE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_right),interface->channels[c].sends[s].label_pan, FALSE, FALSE, FALSE);		                              
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_left),interface->channels[c].sends[s].box_hbox_middle_knob_left, FALSE, TRUE, FALSE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_right),interface->channels[c].sends[s].box_hbox_middle_knob_right, FALSE, TRUE, FALSE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle_knob_left),interface->channels[c].sends[s].knob_gain, TRUE, FALSE, TRUE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle_knob_right),interface->channels[c].sends[s].knob_pan, TRUE, FALSE, TRUE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_left),interface->channels[c].sends[s].box_hbox_middle_entry_left, FALSE, TRUE, FALSE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_right),interface->channels[c].sends[s].box_hbox_middle_entry_right, FALSE, TRUE, FALSE);			                              
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle_entry_left),interface->channels[c].sends[s].entry_gain, TRUE, FALSE, TRUE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle_entry_right),interface->channels[c].sends[s].entry_pan, TRUE, FALSE, TRUE);			                              

				gtk_widget_show (interface->channels[c].sends[s].box_vbox_middle_left);
				gtk_widget_show (interface->channels[c].sends[s].box_vbox_middle_right);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_middle_knob_left);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_middle_knob_right);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_middle_entry_left);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_middle_entry_right);
				
			    gtk_widget_show (interface->channels[c].sends[s].label_pan);
			    gtk_widget_show (interface->channels[c].sends[s].knob_pan);
			    gtk_widget_show (interface->channels[c].sends[s].entry_pan);
			    
			}
			
			if (interface->omit_send_pan && interface->omit_send_gain){
				
				interface->channels[c].sends[s].box_vbox_middle_left = gtk_vbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_middle_knob_left = gtk_hbox_new (FALSE, 0);
				interface->channels[c].sends[s].box_hbox_middle_entry_left = gtk_hbox_new (FALSE, 0);

			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle),interface->channels[c].sends[s].box_vbox_middle_left, TRUE, FALSE, TRUE);			                              
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_left),interface->channels[c].sends[s].label_gain, FALSE, FALSE, FALSE);			                              
				gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_left),interface->channels[c].sends[s].box_hbox_middle_knob_left, FALSE, TRUE, FALSE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle_knob_left),interface->channels[c].sends[s].knob_gain, TRUE, FALSE, TRUE);
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_vbox_middle_left),interface->channels[c].sends[s].box_hbox_middle_entry_left, FALSE, TRUE, FALSE);			                              
			    gtk_box_pack_start (GTK_BOX (interface->channels[c].sends[s].box_hbox_middle_entry_left),interface->channels[c].sends[s].entry_gain, TRUE, FALSE, TRUE);			                              

				gtk_widget_show (interface->channels[c].sends[s].box_vbox_middle_left);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_middle_knob_left);
				gtk_widget_show (interface->channels[c].sends[s].box_hbox_middle_entry_left);
				
			}
			
			gtk_widget_show (interface->channels[c].sends[s].label_gain);
		    gtk_widget_show (interface->channels[c].sends[s].knob_gain);
		    gtk_widget_show (interface->channels[c].sends[s].entry_gain);
		    
		}
		interface->channels[c].separator = gtk_vseparator_new ();
		gtk_box_pack_start (GTK_BOX (interface->box_channels_1_hbox), interface->channels[c].separator, FALSE, FALSE, 0);
		gtk_widget_show (interface->channels[c].separator);
    }
    gtk_widget_show (interface->window_channels_1);
    gtk_widget_show (interface->window_channel_1);
	return 0;
}

static int interface_init_gtk_outputs(jmd_interface_t *interface)
{
	int s;
	char window_title[32];
	char label[32];
	char lc_name[5];
	char *utf_name=NULL;
	
	
    interface->window_output = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    interface->box_outputs_vbox = gtk_vbox_new (FALSE, 0);
	interface->box_outputs_hbox = gtk_hbox_new (FALSE, 0);
    
    interface->file_menu = gtk_menu_new ();    /* Don't need to show menus */
    interface->help_menu = gtk_menu_new ();    /* Don't need to show menus */
    interface->file_item = gtk_menu_item_new_with_label ("File");
    interface->help_item = gtk_menu_item_new_with_label ("Help");
    interface->quit_item = gtk_menu_item_new_with_label ("Quit");
    interface->about_item = gtk_menu_item_new_with_label ("About");
    
    interface->menu_bar = gtk_menu_bar_new ();



    /* Add them to the menu */
    gtk_menu_shell_append (GTK_MENU_SHELL (interface->file_menu), interface->quit_item);
    gtk_menu_shell_append (GTK_MENU_SHELL (interface->help_menu), interface->about_item);
    
    
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (interface->file_item), interface->file_menu);
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (interface->help_item), interface->help_menu);
	
	gtk_menu_bar_append (GTK_MENU_BAR (interface->menu_bar), interface->file_item);
	gtk_menu_bar_append (GTK_MENU_BAR (interface->menu_bar), interface->help_item);
	
	g_signal_connect (G_OBJECT (interface->window_output), "delete_event",
		      G_CALLBACK (gtk_delete), NULL);

    g_signal_connect (G_OBJECT (interface->window_output), "destroy",
		      G_CALLBACK (gtk_destroy), NULL);
    /* Attach the callback functions to the activate signal */
    /* We can attach the Quit menu item to our exit function */
    g_signal_connect_swapped (G_OBJECT (interface->quit_item), "activate",
                              G_CALLBACK (gtk_destroy),
                              (gpointer) "file.quit");
                              
    g_signal_connect (G_OBJECT (interface->about_item), "activate",
                              G_CALLBACK (gtk_about),
                              NULL );

    interface->window_outputs = gtk_scrolled_window_new (NULL, NULL);
    
	sprintf( window_title,"JackMixDesk outs 1- %i",interface->prefader_count+interface->postfader_count);
	gtk_window_set_title (GTK_WINDOW (interface->window_output), window_title);

	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (interface->window_outputs),
                                    GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
                                   
    gtk_container_add (GTK_CONTAINER (interface->window_output),interface->window_outputs );

	gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (interface->window_outputs), interface->box_outputs_vbox);
	
	gtk_box_pack_start (GTK_BOX (interface->box_outputs_vbox),interface->menu_bar, FALSE, FALSE, FALSE);
	gtk_box_pack_start (GTK_BOX (interface->box_outputs_vbox),interface->box_outputs_hbox, TRUE, TRUE, TRUE);
	
    gtk_container_set_border_width (GTK_CONTAINER (interface->window_output), 0);

    gtk_widget_set_size_request (interface->window_output, 110, 580);

    gtk_container_set_border_width (GTK_CONTAINER (interface->window_outputs), 0);

    /* We do need to show menu items */
    gtk_widget_show (interface->box_outputs_vbox);
    gtk_widget_show (interface->quit_item);
    gtk_widget_show (interface->file_item);
    gtk_widget_show (interface->file_menu);
    gtk_widget_show (interface->about_item);
    gtk_widget_show (interface->help_item);
    gtk_widget_show (interface->help_menu);
    gtk_widget_show (interface->menu_bar);
    gtk_widget_show (interface->box_outputs_hbox);

	/* create the widgets for the output window */
    for(s=0;s<(interface->prefader_count+interface->postfader_count);s++){
    	
    	sprintf(label,interface->outputs[s].label);
    	
    	interface->outputs[s].box_vbox = gtk_vbox_new (FALSE, 0);
    	interface->outputs[s].separator =  gtk_vseparator_new ();
    	interface->outputs[s].separator_top =  gtk_hseparator_new ();
    	interface->outputs[s].separator_bottom =  gtk_hseparator_new ();
    	interface->outputs[s].box_hbox_label_top = gtk_hbox_new (FALSE, 0);
    	interface->outputs[s].box_hbox_label_bottom = gtk_hbox_new (FALSE, 0);
    	interface->outputs[s].label_gain = gtk_label_new(label);
    	interface->outputs[s].label_out_bottom = gtk_label_new(label);
    	interface->outputs[s].box_hbox_fader = gtk_hbox_new (FALSE, 0);
    	interface->outputs[s].fader_gain = phat_vfan_slider_new_with_range(interface->outputs[s].gain,KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);
    	interface->outputs[s].box_hbox_entry = gtk_hbox_new (FALSE, 0);
    	interface->outputs[s].entry_gain = gtk_spin_button_new_with_range(KNOB_GAIN_MIN,KNOB_GAIN_MAX,KNOB_GAIN_STEP);
		gtk_widget_set_style(interface->outputs[s].entry_gain,interface->spinbutton_style);
		gtk_spin_button_set_digits(GTK_SPIN_BUTTON(interface->outputs[s].entry_gain),1);  
		interface->outputs[s].box_hbox_mute = gtk_hbox_new (FALSE, 0);
		interface->outputs[s].box_hbox_link = gtk_hbox_new (FALSE, 0);
		interface->outputs[s].button_mute = gtk_check_button_new_with_label("Mute");
    	gtk_widget_set_size_request(interface->outputs[s].box_hbox_link,-1,20);
    	
		if( s < interface->prefader_count + interface->postfader_count -1){
    		interface->outputs[s].label_link = gtk_label_new("Link");
     		interface->outputs[s].button_link = gtk_check_button_new();
			g_signal_connect (G_OBJECT (interface->outputs[s].button_link), "clicked",
					G_CALLBACK (gtk_button_link_out_change),(gpointer) interface );
		}
		
    	gtk_entry_set_width_chars(GTK_ENTRY(interface->outputs[s].entry_gain),DEFAULT_ENTRY_LEN);
    	
    	sprintf(lc_name,"%i",s+1);
		utf_name = stringprep_locale_to_utf8(lc_name);
		gtk_widget_set_name (interface->outputs[s].fader_gain,utf_name);
		gtk_widget_set_name (interface->outputs[s].entry_gain,utf_name);
		gtk_widget_set_name (interface->outputs[s].button_mute,utf_name);
		if( s < interface->prefader_count + interface->postfader_count -1){
			gtk_widget_set_name (interface->outputs[s].button_link,utf_name);
		}
    	
			g_signal_connect (G_OBJECT (interface->outputs[s].fader_gain), "value_changed",
					G_CALLBACK (gtk_fader_gain_out_change),(gpointer) interface );

			g_signal_connect (G_OBJECT (interface->outputs[s].entry_gain), "value_changed",
					G_CALLBACK (gtk_entry_gain_out_change),(gpointer) interface );
					
			g_signal_connect (G_OBJECT (interface->outputs[s].button_mute), "clicked",
					G_CALLBACK (gtk_button_mute_out_change),(gpointer) interface );
    	
    	gtk_box_pack_start (GTK_BOX (interface->box_outputs_hbox),interface->outputs[s].box_vbox, FALSE, FALSE, FALSE);
    	gtk_box_pack_start (GTK_BOX (interface->box_outputs_hbox),interface->outputs[s].separator, FALSE, FALSE, FALSE);
    	gtk_box_pack_start (GTK_BOX (interface->outputs[s].box_vbox),interface->outputs[s].box_hbox_label_top, FALSE, TRUE, FALSE);
    	gtk_box_pack_start (GTK_BOX (interface->outputs[s].box_vbox),interface->outputs[s].separator_top, FALSE, TRUE, FALSE);
    	gtk_box_pack_end (GTK_BOX (interface->outputs[s].box_vbox),interface->outputs[s].box_hbox_label_bottom, FALSE, TRUE, FALSE);
    	gtk_box_pack_end (GTK_BOX (interface->outputs[s].box_vbox),interface->outputs[s].separator_bottom, FALSE, TRUE, FALSE);
		gtk_box_pack_end (GTK_BOX (interface->outputs[s].box_vbox),interface->outputs[s].box_hbox_link, FALSE, TRUE, FALSE);
		gtk_box_pack_end (GTK_BOX (interface->outputs[s].box_vbox),interface->outputs[s].box_hbox_mute, FALSE, TRUE, FALSE);
    	gtk_box_pack_end (GTK_BOX (interface->outputs[s].box_vbox),interface->outputs[s].box_hbox_entry, FALSE, TRUE, FALSE);
    	gtk_box_pack_end (GTK_BOX (interface->outputs[s].box_vbox),interface->outputs[s].box_hbox_fader, TRUE, TRUE, TRUE);
    	gtk_box_pack_end (GTK_BOX (interface->outputs[s].box_hbox_fader),interface->outputs[s].fader_gain, TRUE, FALSE, TRUE);
    	gtk_box_pack_start (GTK_BOX (interface->outputs[s].box_hbox_label_top),interface->outputs[s].label_gain, TRUE, FALSE, TRUE);
    	gtk_box_pack_start (GTK_BOX (interface->outputs[s].box_hbox_label_bottom),interface->outputs[s].label_out_bottom, TRUE, FALSE, TRUE);
    	gtk_box_pack_start (GTK_BOX (interface->outputs[s].box_hbox_mute),interface->outputs[s].button_mute, TRUE, FALSE, TRUE);
    	if( s < interface->prefader_count + interface->postfader_count -1){
    		gtk_box_pack_start (GTK_BOX (interface->outputs[s].box_hbox_link),interface->outputs[s].label_link, TRUE, FALSE, TRUE);
    		gtk_box_pack_start (GTK_BOX (interface->outputs[s].box_hbox_link),interface->outputs[s].button_link, TRUE, FALSE, TRUE);
    	}
    	gtk_box_pack_start (GTK_BOX (interface->outputs[s].box_hbox_entry),interface->outputs[s].entry_gain, TRUE, FALSE, TRUE);
    	
    	gtk_widget_show (interface->outputs[s].box_vbox);
    	gtk_widget_show (interface->outputs[s].separator);
    	gtk_widget_show (interface->outputs[s].separator_top);
    	gtk_widget_show (interface->outputs[s].separator_bottom);
    	gtk_widget_show (interface->outputs[s].box_hbox_label_top);
    	gtk_widget_show (interface->outputs[s].box_hbox_label_bottom);
    	gtk_widget_show (interface->outputs[s].label_gain);
    	gtk_widget_show (interface->outputs[s].label_out_bottom);
    	gtk_widget_show (interface->outputs[s].box_hbox_fader);
    	gtk_widget_show (interface->outputs[s].fader_gain);
    	gtk_widget_show (interface->outputs[s].box_hbox_entry);
    	gtk_widget_show (interface->outputs[s].entry_gain);
    	gtk_widget_show (interface->outputs[s].box_hbox_mute);
    	gtk_widget_show (interface->outputs[s].button_mute);
    	gtk_widget_show (interface->outputs[s].box_hbox_link);
    	if( s < interface->prefader_count + interface->postfader_count -1){
    		gtk_widget_show (interface->outputs[s].label_link);
    		gtk_widget_show (interface->outputs[s].button_link);
    	}
    }

    gtk_widget_show (interface->window_outputs);
    gtk_widget_show (interface->window_output);
    
	return 0;
}

static int interface_update_gtk_channels_1(jmd_interface_t *interface)
{
	int c,s;
	double pan;

	/* update the widgets for the channel window */
    for(c=0;c<interface->channel_count;c++){
    	
		if (interface->channels[c].sends[0].gain_l == 0 && interface->channels[c].sends[0].gain_r < 0){
			pan = interface->channels[c].sends[0].gain_r;
		} else if (interface->channels[c].sends[0].gain_r == 0 && interface->channels[c].sends[0].gain_l < 0){
			pan = interface->channels[c].sends[0].gain_l * (-1);
		} else {
			pan = 0;
		}
		
    	phat_fan_slider_set_value(PHAT_FAN_SLIDER(interface->channels[c].fader_gain),interface->channels[c].sends[0].gain);
	    gtk_spin_button_set_value(GTK_SPIN_BUTTON(interface->channels[c].entry_gain),interface->channels[c].sends[0].gain);
    	
    	if (!interface->omit_chan_pan){
	    	phat_knob_set_value(PHAT_KNOB (interface->channels[c].knob_pan),pan);
	    	gtk_spin_button_set_value(GTK_SPIN_BUTTON(interface->channels[c].entry_pan),pan);
    	}
		/* omit for last channel */
		if( c < interface->channel_count -1){
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(interface->channels[c].button_stereo), interface->channels[c].button_stereo_state);
		}
     	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(interface->channels[c].button_mute),interface->channels[c].button_mute_state);

		for(s=0;s<(interface->prefader_count+interface->postfader_count);s++){
			
			phat_knob_set_value(PHAT_KNOB (interface->channels[c].sends[s].knob_gain),interface->channels[c].sends[s].gain);
			gtk_spin_button_set_value(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_gain),interface->channels[c].sends[s].gain);
			
			if (!interface->omit_send_pan && !interface->omit_send_gain){
				if (interface->channels[c].sends[s].gain_l == 0 && interface->channels[c].sends[s].gain_r < 0){
					pan = interface->channels[c].sends[s].gain_r;
				} else if (interface->channels[c].sends[s].gain_r == 0 && interface->channels[c].sends[s].gain_l < 0){
					pan = interface->channels[c].sends[s].gain_l * (-1);
				} else {
					pan = 0;
				}
				phat_knob_set_value(PHAT_KNOB (interface->channels[c].sends[s].knob_pan),pan);
				phat_knob_set_value(PHAT_KNOB (interface->channels[c].sends[s].knob_gain_l),interface->channels[c].sends[s].gain_l);
				phat_knob_set_value(PHAT_KNOB (interface->channels[c].sends[s].knob_gain_r),interface->channels[c].sends[s].gain_r);
				gtk_spin_button_set_value(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_pan),pan);
				gtk_spin_button_set_value(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_gain_l),interface->channels[c].sends[s].gain_l);
				gtk_spin_button_set_value(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_gain_r),interface->channels[c].sends[s].gain_r);

			}
			
			if (interface->omit_send_pan && !interface->omit_send_gain){
				phat_knob_set_value(PHAT_KNOB (interface->channels[c].sends[s].knob_gain_l),interface->channels[c].sends[s].gain_l);
				phat_knob_set_value(PHAT_KNOB (interface->channels[c].sends[s].knob_gain_r),interface->channels[c].sends[s].gain_r);
				gtk_spin_button_set_value(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_gain_l),interface->channels[c].sends[s].gain_l);
				gtk_spin_button_set_value(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_gain_r),interface->channels[c].sends[s].gain_r);
			}
			
			if (!interface->omit_send_pan && interface->omit_send_gain){
				phat_knob_set_value(PHAT_KNOB (interface->channels[c].sends[s].knob_pan),pan);
				gtk_spin_button_set_value(GTK_SPIN_BUTTON(interface->channels[c].sends[s].entry_pan),pan);
			}
		}
    }
	return 0;
}

static int interface_update_gtk_outputs(jmd_interface_t *interface)
{
	int s;

	/* update the widgets for the output window */
    for(s=0;s<(interface->prefader_count+interface->postfader_count);s++){
 
		phat_fan_slider_set_value(PHAT_FAN_SLIDER(interface->outputs[s].fader_gain),interface->outputs[s].gain);
    	gtk_spin_button_set_value(GTK_SPIN_BUTTON(interface->outputs[s].entry_gain),interface->outputs[s].gain);
     	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(interface->outputs[s].button_mute), interface->outputs[s].button_mute_state);
    	
		if( s < interface->prefader_count + interface->postfader_count -1){
     		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(interface->outputs[s].button_link), interface->outputs[s].button_link_state);

		}
	}
	return 0;
}

void *interface_refresh(void *ptr)
{
	jmd_interface_t *interface = (jmd_interface_t*)ptr;
	
	while(running){
		if ( lo_send_from(interface->t,interface->serv,LO_TT_IMMEDIATE, "/mixer/changed_state_get", "") == -1) {
			printf("OSC error %d: %s\n", lo_address_errno(interface->t), lo_address_errstr(interface->t));
		}
		sleep(5);
	}
	return 0;
}

static void interface_finish(jmd_interface_t *interface)
{
	if (interface->verbose) printf( "Closed gtk interface.\n");
	
	free(interface->channels);
	free(interface->outputs);
	free(interface);
}

static int osc_generic_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
	/* catch any incoming messages and display them. returning 1 means that the
	 * message has not been fully handled and the server should try other methods */
    int i;
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		if (interface->verbose>1) {
	    	printf("\n%s received unknown OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
		    for (i=0; i<argc; i++) {
				printf("arg %d '%c' ", i, types[i]);
				lo_arg_pp(types[i], argv[i]);
				printf("\n");
		    }
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 1;
}

static int osc_set_return_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
	/* handle all set returns */
    int i;
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		if (interface->verbose) {
	    	printf("\n%s received set return OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
		    for (i=0; i<argc; i++) {
				printf("arg %d '%c' ", i, types[i]);
				lo_arg_pp(types[i], argv[i]);
				printf("\n");
		    }
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 0;
}

static int osc_channel_count_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		interface->channel_count=argv[0]->i;
		if (interface->verbose>1) {
	    	printf("\n%s received channel count OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
			printf("arg '%c' ", types[0]);
			lo_arg_pp(types[0], argv[0]);
			printf("\n");
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 0;
}

static int osc_prefader_count_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		interface->prefader_count=argv[0]->i;
		if (interface->verbose>1) {
	    	printf("\n%s received prefader count OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
			printf("arg '%c' ", types[0]);
			lo_arg_pp(types[0], argv[0]);
			printf("\n");
		    fflush(stdout); 
		}
	} else {
		return -1;
	}
    return 0;
}

static int osc_postfader_count_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		interface->postfader_count=argv[0]->i;
		if (interface->verbose>1) {
	    	printf("\n%s received postfader count OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
			printf("arg '%c' ", types[0]);
			lo_arg_pp(types[0], argv[0]);
			printf("\n");
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}	
    return 0;
}

static int osc_changed_state_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		if(argv[0]->i){
			interface_refresh_channels(interface);
			interface_init_outputs_data(interface);
		}
		if (interface->verbose) {
	    	printf("\n%s received changed state OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
			printf("arg '%c' ", types[0]);
			lo_arg_pp(types[0], argv[0]);
			printf("\n");
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}	
    return 0;
}

static int osc_channel_label_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    int i;
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		snprintf( interface->channels[(argv[0]->i)-1].label, DEFAULT_LABEL_LEN,"%s",argv[1]->m );
		if (interface->verbose>1) {
	    	printf("\n%s received channel label OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
		    for (i=0; i<argc; i++) {
				printf("arg %d '%c' ", i, types[i]);
				lo_arg_pp(types[i], argv[i]);
				printf("\n");
		    }
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 0;
}

static int osc_channel_mute_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    int i;
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		interface->channels[(argv[0]->i)-1].button_mute_state=argv[1]->i;
		if (interface->verbose) {
	    	printf("\n%s received channel mute OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
		    for (i=0; i<argc; i++) {
				printf("arg %d '%c' ", i, types[i]);
				lo_arg_pp(types[i], argv[i]);
				printf("\n");
		    }
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 0;
}

static int osc_channel_stereo_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    int i;
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		interface->channels[(argv[0]->i)-1].button_stereo_state=argv[1]->i;
		if (interface->verbose) {
	    	printf("\n%s received channel stereo OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
		    for (i=0; i<argc; i++) {
				printf("arg %d '%c' ", i, types[i]);
				lo_arg_pp(types[i], argv[i]);
				printf("\n");
		    }
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 0;
}

static int osc_channel_send_gain_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    int i;
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		interface->channels[(argv[0]->i)-1].sends[(argv[1]->i)-1].gain=argv[2]->f;
		if (interface->verbose>1) {
	    	printf("\n%s received channel send gain OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
		    for (i=0; i<argc; i++) {
				printf("arg %d '%c' ", i, types[i]);
				lo_arg_pp(types[i], argv[i]);
				printf("\n");
		    }
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 0;
}

static int osc_channel_send_pan_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    int i;
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		interface->channels[(argv[0]->i)-1].sends[(argv[1]->i)-1].pan=argv[2]->f;
		if (interface->verbose>1) {
	    	printf("\n%s received channel send pan OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
		    for (i=0; i<argc; i++) {
				printf("arg %d '%c' ", i, types[i]);
				lo_arg_pp(types[i], argv[i]);
				printf("\n");
		    }
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 0;
}

static int osc_channel_send_gain_l_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    int i;
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		interface->channels[(argv[0]->i)-1].sends[(argv[1]->i)-1].gain_l=argv[2]->f;
		if (interface->verbose>1) {
	    	printf("\n%s received channel send gain left OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
		    for (i=0; i<argc; i++) {
				printf("arg %d '%c' ", i, types[i]);
				lo_arg_pp(types[i], argv[i]);
				printf("\n");
		    }
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 0;
}

static int osc_channel_send_gain_r_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    int i;
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {

		interface->channels[(argv[0]->i)-1].sends[(argv[1]->i)-1].gain_r=argv[2]->f;
		if (interface->verbose>1) {
	    	printf("\n%s received channel send gain right OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
		    for (i=0; i<argc; i++) {
				printf("arg %d '%c' ", i, types[i]);
				lo_arg_pp(types[i], argv[i]);
				printf("\n");
		    }
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 0;
}

static int osc_output_label_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    int i;
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		snprintf( interface->outputs[(argv[0]->i)-1].label, DEFAULT_LABEL_LEN,"%s",argv[1]->m );
		for(i=0;i<interface->channel_count;i++){
			snprintf( interface->channels[i].sends[(argv[0]->i)-1].label, DEFAULT_LABEL_LEN,"%s",argv[1]->m );
		}
		if (interface->verbose>1) {
	    	printf("\n%s received output label OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
		    for (i=0; i<argc; i++) {
				printf("arg %d '%c' ", i, types[i]);
				lo_arg_pp(types[i], argv[i]);
				printf("\n");
		    }
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 0;
}

static int osc_output_gain_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    int i;
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		interface->outputs[(argv[0]->i)-1].gain=argv[1]->f;
		if ((interface->verbose)>1) {
	    	printf("\n%s received output gain OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
		    for (i=0; i<argc; i++) {
				printf("arg %d '%c' ", i, types[i]);
				lo_arg_pp(types[i], argv[i]);
				printf("\n");
		    }
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 0;
}

static int osc_output_mute_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    int i;
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		interface->outputs[(argv[0]->i)-1].button_mute_state=argv[1]->i;
		if (interface->verbose) {
	    	printf("\n%s received output mute OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
		    for (i=0; i<argc; i++) {
				printf("arg %d '%c' ", i, types[i]);
				lo_arg_pp(types[i], argv[i]);
				printf("\n");
		    }
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 0;
}

static int osc_output_link_handler(const char *path, const char *types, lo_arg **argv,
			int argc, void *data, void *user_data)
{
    int i;
    jmd_interface_t *interface = (jmd_interface_t*)user_data;
	if (data) {
		interface->outputs[(argv[0]->i)-1].button_link_state=argv[1]->i;
		if (interface->verbose) {
	    	printf("\n%s received output link OSC message.\n",interface->client_name);
		    printf("path: <%s>\n", path);
		    for (i=0; i<argc; i++) {
				printf("arg %d '%c' ", i, types[i]);
				lo_arg_pp(types[i], argv[i]);
				printf("\n");
		    }
		    fflush(stdout); 
		}   	
	} else {
		return -1;
	}
    return 0;
}

static void osc_server_error(int num, const char *msg, const char *path)
{
    printf("liblo server error %d in path %s: %s\n", num, path, msg);
}

static void osc_server_init(jmd_interface_t *interface) 
{
	/* Create new server */
	interface->st = lo_server_thread_new(NULL, osc_server_error );

	/* Add the methods */
    lo_server_thread_add_method(interface->st, "/mixer/channel_count_is", "i", osc_channel_count_handler,interface);
    lo_server_thread_add_method(interface->st, "/mixer/prefader_count_is", "i", osc_prefader_count_handler,interface);
    lo_server_thread_add_method(interface->st, "/mixer/postfader_count_is", "i", osc_postfader_count_handler,interface);
    lo_server_thread_add_method(interface->st, "/mixer/changed_state_is", "i", osc_changed_state_handler,interface);
    
    lo_server_thread_add_method(interface->st, "/mixer/channel/label_is", "is", osc_channel_label_handler,interface);
    lo_server_thread_add_method(interface->st, "/mixer/channel/send/gain_is", "iif", osc_channel_send_gain_handler, interface);
    lo_server_thread_add_method(interface->st, "/mixer/channel/send/pan_is", "iif", osc_channel_send_pan_handler, interface);
    lo_server_thread_add_method(interface->st, "/mixer/channel/send/gain_l_is", "iif", osc_channel_send_gain_l_handler, interface);
    lo_server_thread_add_method(interface->st, "/mixer/channel/send/gain_r_is", "iif", osc_channel_send_gain_r_handler, interface);
    lo_server_thread_add_method(interface->st, "/mixer/channel/mute_is", "ii", osc_channel_mute_handler,interface);
    lo_server_thread_add_method(interface->st, "/mixer/channel/stereo_is", "ii", osc_channel_stereo_handler,interface);
    
    lo_server_thread_add_method(interface->st, "/mixer/output/label_is", "is", osc_output_label_handler,interface);
    lo_server_thread_add_method(interface->st, "/mixer/output/gain_is", "if", osc_output_gain_handler, interface);
    lo_server_thread_add_method(interface->st, "/mixer/output/mute_is", "ii", osc_output_mute_handler,interface);
    lo_server_thread_add_method(interface->st, "/mixer/output/link_is", "ii", osc_output_link_handler,interface);

    lo_server_thread_add_method(interface->st, "/mixer/channel/label_now", "is", osc_set_return_handler,interface);
    lo_server_thread_add_method(interface->st, "/mixer/channel/send/gain_now", "iif", osc_set_return_handler, interface);
    lo_server_thread_add_method(interface->st, "/mixer/channel/send/pan_now", "iif", osc_set_return_handler, interface);
    lo_server_thread_add_method(interface->st, "/mixer/channel/send/gain_l_now", "iif", osc_set_return_handler, interface);
    lo_server_thread_add_method(interface->st, "/mixer/channel/send/gain_r_now", "iif", osc_set_return_handler, interface);
    lo_server_thread_add_method(interface->st, "/mixer/channel/mute_now", "ii", osc_set_return_handler,interface);
    lo_server_thread_add_method(interface->st, "/mixer/channel/stereo_now", "ii", osc_set_return_handler,interface);
    
    lo_server_thread_add_method(interface->st, "/mixer/output/label_now", "is", osc_set_return_handler,interface);
    lo_server_thread_add_method(interface->st, "/mixer/output/gain_now", "if", osc_set_return_handler, interface);
    lo_server_thread_add_method(interface->st, "/mixer/output/mute_now", "ii", osc_set_return_handler,interface);
    lo_server_thread_add_method(interface->st, "/mixer/output/link_now", "ii", osc_set_return_handler,interface);

    /* add method that will match any path and args got to be last*/
    lo_server_thread_add_method(interface->st, NULL, NULL, osc_generic_handler,interface);

    lo_server_thread_start(interface->st);
}

static void osc_server_finish(jmd_interface_t *interface)
{
	lo_server_thread_stop(interface->st);	lo_server_thread_free(interface->st);
	
	if (interface->verbose) printf( "Stopped OSC server thread.\n");
}

int main(int argc, char *argv[])
{
	int opt;
	/*pthread_t refresh_thread;*/
	
	jmd_interface_t *interface = NULL;
	
    gtk_init (&argc, &argv);
    
    interface=interface_init();
    
	while ((opt = getopt(argc, argv, "n:p:m:dgcvqh")) != -1) {
		switch (opt) {
			case 'n':  snprintf( interface->client_name, DEFAULT_LABEL_LEN, "%s", optarg ); break;
			case 'v':  interface->verbose=1; break;
			case 'q':  interface->quiet=1; break;
			case 'd':  interface->omit_send_pan=1; break;
			case 'g':  interface->omit_send_gain=1; break;
			case 'c':  interface->omit_chan_pan=1; break;
			case 'p':  interface->osc_mixer_port = optarg; break;
			case 'm':  interface->osc_mixer_host = optarg; break;
			default:
				fprintf(stderr, "Unknown option '%c'.\n", (char)opt);
			case 'h':
				interface_usage( argv[0] );
				break;
		}
	}
    argc -= optind;
    argv += optind;
    
	osc_server_init(interface);
	
	interface_init_mixer(interface);
	
	/* wait for osc server and/or client to finish starting */
    sleep(1);
    
    if (!interface->quiet) {
		char *url = lo_server_thread_get_url(interface->st);
		printf( "%s\nOSC server URL: %s\n", interface->client_name, url );
		free(url);
	}
	
	/* Dislay welcoming message */
	if (interface->verbose) {
		printf("\nStarting %s with %i channels %i prefaders %i postfaders.\n",
			interface->client_name, interface->channel_count,interface->prefader_count,interface->postfader_count);
		printf("JackMixDesk OSC Port: %s\n",interface->osc_mixer_port);
	}

    interface_init_channels(interface);
    
    interface_init_outputs_struct(interface);
 
    interface_init_outputs_data(interface);
        
    interface_init_gtk_outputs(interface);
    
    interface_init_gtk_channels_1(interface);
    
    interface_update_gtk_outputs(interface);
    
    interface_update_gtk_channels_1(interface);
    
	/*pthread_create( &refresh_thread, NULL, interface_refresh, interface );*/
	
    gtk_main ();
    
    osc_server_finish(interface);
    
    interface_finish(interface);
    
    return 0;
}
